# plugins/loader.py
import importlib.util
import os
import inspect
from roibaview.plugins.base import BasePlugin


def load_plugins(context=None):
    plugin_dir = os.path.dirname(__file__)
    plugins = []

    for fname in os.listdir(plugin_dir):
        if fname.endswith(".py") and fname not in {"__init__.py", "base.py", "loader.py"}:
            path = os.path.join(plugin_dir, fname)
            spec = importlib.util.spec_from_file_location(fname[:-3], path)
            module = importlib.util.module_from_spec(spec)
            try:
                spec.loader.exec_module(module)
            except Exception as e:
                print(f"Error loading plugin {fname}: {e}")
                continue

            # Find plugin classes
            for _, obj in inspect.getmembers(module, inspect.isclass):
                if issubclass(obj, BasePlugin) and obj is not BasePlugin:
                    try:
                        try:
                            plugin = obj(**context) if context else obj()
                            plugins.append(plugin)
                        except Exception as e:
                            print(f"Failed to instantiate plugin {obj.__name__}: {e}")
                    except Exception as e:
                        print(f"Failed to instantiate plugin {obj.__name__}: {e}")
    return plugins



from PyQt6.QtGui import QFont
from PyQt6.QtCore import pyqtSignal, Qt, QEvent
from PyQt6.QtWidgets import QMainWindow, QPushButton, QWidget, QLabel, QVBoxLayout, \
    QMessageBox, QHBoxLayout, QComboBox, QToolBar, QListWidget, \
    QScrollArea, QMenu
import pyqtgraph as pg


class MainWindow(QMainWindow):
    # PyQT Signals
    key_modifier_pressed = pyqtSignal(int)
    key_pressed = pyqtSignal(QEvent)
    key_released = pyqtSignal(QEvent)
    # closing = pyqtSignal()

    def __init__(self, screen):
        super().__init__()
        self.screen = screen

        # Setup GUI Elements
        self._setup_ui()

        # Set Window Size
        # self.resize(800, 800)
        screen_h = self.screen.height()
        screen_w = self.screen.width()
        # self.showMaximized()
        # setGeometry(left, top, width, height)
        window_width = 1000
        window_height = 800
        self.setGeometry(screen_w // 2 - window_width // 2, screen_h // 2 - window_height // 2, window_width,
                         window_height)

    def _setup_ui(self):
        self.setWindowTitle("Ca Event Analysis")
        # Central Widget of the Main Window
        self.centralWidget = QWidget()
        self.setCentralWidget(self.centralWidget)
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Create Widgets
        # Create Plotting Area
        self.plot_graphics_layout_widget = pg.GraphicsLayoutWidget(show=False, title='')

        # Data Plot
        self.trace_plot_item = self.plot_graphics_layout_widget.addPlot(title='', clear=True, name='data')
        self.trace_plot_item.hideButtons()

        # Dataset List
        self.data_sets_list = QListWidget()
        self.data_sets_list.setWindowTitle('Data sets')
        self.data_sets_list.setSelectionMode(QListWidget.SelectionMode.MultiSelection)  # Set selection mode to multi-selection
        # self.data_sets_list.addItems(["DataSet_" + str(i) for i in range(50)])  # Adding some items for demonstration
        self.data_sets_list.setMaximumWidth(100)

        # Set stylesheet for selected and unselected items
        self.data_sets_list.setStyleSheet(
            "QListWidget::item:selected {background-color: blue; color: white;}"
            "QListWidget::item:selected:!active {background-color: lightblue; color: white}"  # not active
            "QListWidget::item:!selected {background-color: white; color: black;}")

        # Create scroll area and set list widget as its widget
        self.data_sets_list_scroll_area = QScrollArea()
        self.data_sets_list_scroll_area.setWidget(self.data_sets_list)
        self.data_sets_list_scroll_area.setWidgetResizable(False)  # Allow the scroll area to resize its contents

        # Create context menu for data list
        self.data_sets_list_context_menu = QMenu()
        self.data_sets_list_rename = self.data_sets_list_context_menu.addAction("rename")
        self.data_sets_list_delete = self.data_sets_list_context_menu.addAction("delete")
        self.data_sets_list_delete_col = self.data_sets_list_context_menu.addAction("delete column")

        self.data_sets_list_export = self.data_sets_list_context_menu.addAction("export")

        self.data_sets_list_context_menu.addSeparator()
        self.data_sets_list_time_offset = self.data_sets_list_context_menu.addAction("time offset")
        self.data_sets_list_y_offset = self.data_sets_list_context_menu.addAction("y offset")

        self.data_sets_list_context_menu.addSeparator()
        self.style_menu = self.data_sets_list_context_menu.addMenu('Style')
        self.style_color = self.style_menu.addAction("Change Color")
        self.style_lw = self.style_menu.addAction("Line Width")

        # Connect right-click event to show context menu
        self.data_sets_list.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.data_sets_list.customContextMenuRequested.connect(self.show_context_menu)

        # The Toolbar
        self.toolbar = QToolBar()
        self.toolbar.setFloatable(False)
        self.toolbar.setMovable(False)
        self.toolbar.toggleViewAction().setEnabled(False)
        self.addToolBar(self.toolbar)

        # The Mouse Position
        self.layout_labels = QHBoxLayout()
        self.mouse_label = QLabel(f"<p style='color:black'>X： {0} <br> Y: {0}</p>")

        # Info Label
        self.info_label = QLabel('')
        self.info_frame_rate = QLabel('')

        # ROI Selection Drop Down
        self.roi_selection_combobox_label = QLabel('ROI: ')
        self.roi_selection_combobox = QComboBox()

        # Create Layout
        self.layout_labels.addWidget(self.mouse_label)
        self.layout_labels.addStretch()
        self.layout_labels.addWidget(self.info_label)
        self.layout_labels.addStretch()
        self.layout_labels.addWidget(self.info_frame_rate)
        self.layout_labels.addStretch()
        self.layout_labels.addWidget(self.roi_selection_combobox_label)
        self.layout_labels.addWidget(self.roi_selection_combobox)
        self.layout_labels.addStretch()

        # buttons
        button_text_font = QFont('Sans Serif', 12)
        button_text_font.setBold(True)
        self.next_button = QPushButton('>>', self)
        self.prev_button = QPushButton('<<', self)
        self.next_button.setFont(button_text_font)
        self.prev_button.setFont(button_text_font)

        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Set Button Layout
        layout_buttons = QHBoxLayout()
        layout_buttons.addWidget(self.prev_button)
        layout_buttons.addWidget(self.next_button)

        # Set Main Layout
        # Vertical Box Layout
        layout = QVBoxLayout()
        layout_center_window = QHBoxLayout()
        layout_center_window.addWidget(self.data_sets_list)
        layout_center_window.addWidget(self.plot_graphics_layout_widget, stretch=1)

        # Add the widgets to the layout
        layout.addLayout(self.layout_labels)
        layout.addLayout(layout_center_window)
        layout.addLayout(layout_buttons)
        layout.setStretchFactor(self.plot_graphics_layout_widget, 4)
        # Connect the layout to the central widget
        self.centralWidget.setLayout(layout)

        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # File Menu
        self.menu = self.menuBar()
        self.file_menu = self.menu.addMenu("&File")
        self.file_menu_new_viewer_file = self.file_menu.addAction('New ... (ctrl+n)')
        self.file_men_open_viewer_file = self.file_menu.addAction('Open ... (ctrl+o)')
        self.file_menu_save_viewer_file = self.file_menu.addAction('Save Viewer File (ctrl+s)')
        self.file_menu.addSeparator()
        self.file_menu_import_csv = self.file_menu.addAction('Import csv file ...')

        self.file_menu.addSeparator()
        self.file_menu_action_exit = self.file_menu.addAction('Exit')

        # Tools Menu
        self.tools_menu = self.menu.addMenu('Tools')
        self.tools_menu_open_video_viewer = self.tools_menu.addAction('Open Video Viewer')
        self.tools_menu_convert_ventral_root = self.tools_menu.addAction('Convert Ventral Root Files')
        self.tools_menu_create_stimulus = self.tools_menu.addAction('Create Stimulus From File')

        # Utils Menu
        self.utils_menu = self.menu.addMenu('Utils')

        # PLUGINS
        # Filter
        self.plugins_filter_menu = self.data_sets_list_context_menu.addMenu('Filter')
        self.plugins_filter_menu_actions = []  # Keep track of actions dynamically

        # Transformations
        self.plugins_transformation_menu = self.data_sets_list_context_menu.addMenu('Transformation')
        self.plugins_transformation_menu_actions = []  # Keep track of actions dynamically

    def populate_transformation_plugins_menu(self, plugins, callback):
        self.plugins_transformation_menu.clear()
        for plugin in plugins:
            action = self.plugins_transformation_menu.addAction(plugin.name)
            action.triggered.connect(lambda _, p=plugin: callback(p))

    def populate_filter_plugins_menu(self, plugins, callback):
        self.plugins_filter_menu.clear()
        for plugin in plugins:
            action = self.plugins_filter_menu.addAction(plugin.name)
            action.triggered.connect(lambda _, p=plugin: callback(p))

    def add_tools_menu_plugins(self, plugins):
        """
        Add all tool plugins to the Tools menu.
        Disabled ones are visible but grayed out.
        """
        for plugin in plugins:
            action = self.tools_menu.addAction(plugin.name)

            # Assign shortcut if plugin defines one
            if hasattr(plugin, "shortcut"):
                action.setShortcut(plugin.shortcut)

            if not plugin.available():
                action.setEnabled(False)
                reason = getattr(plugin, "unavailable_reason", "Plugin not available.")
                action.setToolTip(reason)
            else:
                action.triggered.connect(plugin.apply)

    def add_utils_menu_plugins(self, plugins):
        """
        Add all tool plugins to the Utils menu.
        Disabled ones are visible but grayed out.
        """
        for plugin in plugins:
            action = self.utils_menu.addAction(plugin.name)

            # Assign shortcut if plugin defines one
            if hasattr(plugin, "shortcut"):
                action.setShortcut(plugin.shortcut)

            if not plugin.available():
                action.setEnabled(False)
                reason = getattr(plugin, "unavailable_reason", "Plugin not available.")
                action.setToolTip(reason)
            else:
                action.triggered.connect(plugin.apply)

    def show_context_menu(self, pos):
        # Show context menu at the position of the mouse cursor
        self.data_sets_list_context_menu.exec(self.data_sets_list.mapToGlobal(pos))

    def keyPressEvent(self, event):
        super(MainWindow, self).keyPressEvent(event)
        self.key_pressed.emit(event)

    def keyReleaseEvent(self, event):
        super(MainWindow, self).keyReleaseEvent(event)
        self.key_released.emit(event)

    def freeze_gui(self, freeze=True):
        self.data_sets_list.setDisabled(freeze)

    @staticmethod
    def exit_dialog():
        msg_box = QMessageBox()
        msg_box.setText('Exit ...')
        msg_box.setInformativeText('Do you want to save your changes?')
        msg_box.setStandardButtons(
            QMessageBox.StandardButton.Save | QMessageBox.StandardButton.Discard | QMessageBox.StandardButton.Cancel)
        msg_box.setDefaultButton(QMessageBox.StandardButton.Save)
        retval = msg_box.exec()
        return retval


import pyqtgraph as pg
import numpy as np
from PyQt6.QtCore import pyqtSignal, Qt
from PyQt6.QtWidgets import QVBoxLayout, QSlider, QLabel, QSpacerItem, QSizePolicy, QMessageBox, QDialog, QPushButton
from scipy import signal
from roibaview.gui_utils import BrowseFileDialog
import pandas as pd
from roibaview.plugins.base import BasePlugin


class PeakDetectionPlugin(BasePlugin):
    name = "Peak Detection"
    category = "tool"
    shortcut = "Ctrl+Shift+P"

    def __init__(self, config=None, parent=None):
        self.config = config
        self.parent = parent  # mostly the main gui
        self.dialog = None

    def apply(self, *_):
        try:
            # Get the controller (connected to the gui)
            controller = getattr(self.parent, "controller", None)
            if not controller:
                raise RuntimeError("Controller not found in GUI context.")
            if not controller.selected_data_sets:
                raise RuntimeError("No dataset selected.")

            name = controller.selected_data_sets[0]
            dtype = controller.selected_data_sets_type[0]
            roi = controller.current_roi_idx
            data = controller.data_handler.get_data_set(dtype, name)
            meta = controller.data_handler.get_data_set_meta_data(dtype, name)
            fr = meta["sampling_rate"]
            master_plot = controller.data_plotter.master_plot

            self.dialog = PeakDetection(
                data=data + meta["y_offset"],
                fr=fr,
                master_plot=master_plot,
                roi=roi,
                parent=self.parent,
                controller=controller
            )

            # Connect signal and cleanup handler
            controller.signal_roi_idx_changed.connect(self.dialog.roi_changed)
            self.dialog.finished.connect(self._on_dialog_closed)  # cleanup

            self.dialog.show()

        except Exception as e:
            from PyQt6.QtWidgets import QMessageBox
            QMessageBox.critical(self.parent, "Peak Detection Error", str(e))

    def _on_dialog_closed(self):
        self.dialog.deleteLater()
        self.dialog = None


# class PeakDetection(QWidget):
class PeakDetection(QDialog):

    signal_roi_changed = pyqtSignal(int)
    main_window_closing = pyqtSignal()

    def __init__(self, data, fr, master_plot, roi, parent=None, controller=None, **kwargs):
        super().__init__(parent)
        self.controller = controller  # Store controller
        self.roi_idx = roi
        self.data = data  # this is the data set
        self.data_trace = self.data[:, self.roi_idx]  # this is the roi trace
        self.fr = fr
        self.time_axis = self.compute_time_axis(self.data_trace.shape[0], self.fr)
        self.master_plot = master_plot
        self.parameters = dict()
        self.parameters_range = dict()
        self.peaks = dict()
        self.set_parameters()
        self.min_range = 0
        self.max_range = 1000
        # self.set_parameters_range()
        self._init_ui()
        self.main_window_running = True
        self.signal_roi_changed.connect(self.roi_changed)
        self.main_window_closing.connect(self.main_window_is_closing)

    def export_peaks(self):
        file_browser = BrowseFileDialog(self)
        file_dir = file_browser.save_file_name('csv file, (*.csv)')
        if file_dir:
            result = pd.DataFrame()
            result['Time'] = self.peaks['times']
            result['ID'] = self.peaks['idx']
            for k in self.peaks['props']:
                result[k] = self.peaks['props'][k]
            result.to_csv(file_dir)

    @staticmethod
    def compute_time_axis(data_size, fr):
        max_time = data_size / fr
        return np.linspace(0, max_time, data_size)

    def main_window_is_closing(self):
        self.main_window_running = False
        self.close()

    def roi_changed(self):
        if self.controller:
            self.roi_idx = self.controller.current_roi_idx
            self.data_trace = self.data[:, self.roi_idx]
            self.peaks['times'], self.peaks['idx'], self.peaks['props'] = self.find_peaks(parameters=self.parameters)
            self.update_plot()

    def _init_ui(self):
        layout = QVBoxLayout()
        self.parameters_labels = dict()
        for param_name in self.parameters.keys():
            label = QLabel(param_name)
            slider = QSlider(Qt.Orientation.Horizontal)
            slider.setMinimum(self.min_range)
            slider.setMaximum(self.max_range)
            slider.setValue(0)  # Initial value
            slider.setTickInterval(1)
            slider.setTickPosition(QSlider.TickPosition.TicksBelow)
            slider.setTickPosition(QSlider.TickPosition.NoTicks)

            self.parameters_labels[param_name] = QLabel('0')

            # Connect slider signal to update function
            slider.valueChanged.connect(lambda value, param=param_name: self.update_parameter(param, value))

            layout.addWidget(label)
            layout.addWidget(slider)
            layout.addWidget(self.parameters_labels[param_name])
            # Add spacer (width, height)
            spacer = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
            layout.addItem(spacer)

        self.export_button = QPushButton('Export...')
        self.export_button.clicked.connect(self.export_peaks)
        layout.addWidget(self.export_button)

        self.setLayout(layout)
        self.setWindowTitle("Peak Detection Parameters")
        # self.show()

    def update_parameter(self, param_name, value):
        # Update parameter value
        if value == 0:
            self.parameters[param_name] = None
        else:
            self.parameters[param_name] = value

        parameters = self.parameters
        if value is not None:
            if param_name == 'height':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 0, np.max(self.data_trace))
            if param_name == 'threshold':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 0, np.max(self.data_trace)*0.1)
            if param_name == 'prominence':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 0, np.max(self.data_trace)*0.25)
            if param_name == 'distance':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 1, self.data_trace.shape[0]*0.5)
            if param_name == 'width':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 1, self.data_trace.shape[0]*0.05)
        else:
            parameters[param_name] = None

        # Update Value Label
        self.parameters_labels[param_name].setText(f'{parameters[param_name]:.2f}')

        # Update find peaks
        self.peaks['times'], self.peaks['idx'], self.peaks['props'] = self.find_peaks(parameters)

        # Update Plot
        self.update_plot()

    def clear_plot(self):
        for item in self.master_plot.items[:]:
            name = item.name()
            # print(f"Checking item: {name}")
            if name and name.startswith("peaks_roi_"):
                # print(f"Removing item: {name}")
                self.master_plot.removeItem(item)

    def update_plot(self):
        # Check if there is already roi data plotted and remove it
        self.clear_plot()

        # Create new  plot item
        plot_data_item = pg.ScatterPlotItem(
            # self.peaks['times'], self.data_trace[self.peaks['idx']],
            self.time_axis[self.peaks['idx']], self.data_trace[self.peaks['idx']],
            pen=pg.mkPen(color=(255, 0, 0)),
            brush=pg.mkBrush(color=(255, 0, 0)),
            size=50,
            symbol='arrow_down',
            # name=f'peaks',
            name=f'peaks_roi_{self.roi_idx}',  # unique name
            skipFiniteCheck=True,
            tip=None,
            hoverable=True,
            hoverSize=100
        )
        # Add plot item to the plot widget
        self.master_plot.addItem(plot_data_item)

    def set_parameters(self):
        self.parameters['height'] = None
        self.parameters['threshold'] = None
        self.parameters['distance'] = None
        self.parameters['prominence'] = None
        # self.parameters['prominence_wlen'] = None
        self.parameters['width'] = None
        # self.parameters['width_rel_height'] = None
        # self.parameters['plateau_size'] = None

    @staticmethod
    def map_range(value, from_min, from_max, to_min, to_max):
        # First, scale the value from the input range to a 0-1 range
        scaled_value = (value - from_min) / (from_max - from_min)

        # Then, scale the value to fit within the output range
        mapped_value = to_min + (to_max - to_min) * scaled_value

        return mapped_value

    def find_peaks(self, parameters):
        peaks, props = signal.find_peaks(self.data_trace, **parameters)
        peaks_time = peaks / self.fr
        return peaks_time, peaks, props

    def closeEvent(self, event):
        if self.main_window_running:
            reply = QMessageBox.question(
                self, 'Message',
                "Are you sure to quit?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No
            )
            if reply == QMessageBox.StandardButton.Yes:
                self._cleanup_on_close()
                self.done(QDialog.DialogCode.Accepted)
                event.accept()
            else:
                event.ignore()
        else:
            self._cleanup_on_close()
            self.done(QDialog.DialogCode.Accepted)
            event.accept()

    def _cleanup_on_close(self):
        """Disconnect signals, remove peak markers, and prepare dialog for deletion."""
        # Disconnect internal signals
        try:
            self.signal_roi_changed.disconnect(self.roi_changed)
        except Exception:
            pass

        try:
            self.main_window_closing.disconnect()
        except Exception:
            pass

        # Disconnect controller signal
        if hasattr(self, "controller") and self.controller:
            try:
                self.controller.signal_roi_idx_changed.disconnect(self.roi_changed)
            except Exception:
                pass

        # Remove any plotted peak markers
        self.clear_plot()

    # def __del__(self):
    #     print("PLUGIN dialog deleted.")



import os
import numpy as np
import pandas as pd
import pickle


def pickle_stuff(file_name, data=None):
    if not file_name.endswith('.pickle'):
        print('ERROR: File name must end with ".pickle"')
        return None
    if data is None:
        with open(file_name, 'rb') as handle:
            result = pickle.load(handle)
        return result
    else:
        with open(file_name, 'wb') as handle:
            pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)
        return True


def convert_time_stamps_to_secs(data, method=0):
    if method == 0:
        # INPUT: hhmmss.ms (163417.4532)
        s = str(data)
        in_secs = int(s[:2]) * 3600 + int(s[2:4]) * 60 + float(s[4:])
    elif method == 1:
        # INPUT: hh:mm:ss
        s = data
        in_secs = int(s[:2]) * 3600 + int(s[3:5]) * 60 + float(s[6:])
    else:
        in_secs = None
    return in_secs


def transform_ventral_root_recording(vr_files, vr_fr, first_file_missing=0):
    # vr_files is a list of ventral root recording text files of one sweep
    vr_files = list(np.sort(vr_files))

    # Get First Recording Text File
    f_name_01 = vr_files[0]

    # Load Text File
    dummy_01 = pd.read_csv(f_name_01, sep='\t', header=None)

    # Get the Voltage values
    vr_values = dummy_01.iloc[:, 0].to_numpy()

    # Get end time point for finding the gap to the next recording file
    t_last = convert_time_stamps_to_secs(dummy_01.iloc[-1, 3], method=0)

    # Loop over the rest of the vr recording text files
    for f_name in vr_files[1:]:
        # check file size
        if os.path.getsize(f_name) <= 10:
            print('')
            print('WARNING')
            print(f'{f_name}: File Size is too small. Something is wrong with this file. Please check!')
            print('Will skip this file and set all values to zero')
            print('')
            continue

        # Load ventral root text file
        dummy = pd.read_csv(f_name, sep='\t', header=None)

        # Compute time distance between the end of the last vr recording and the start of this one (in seconds)
        t_rel_distance = convert_time_stamps_to_secs(dummy.iloc[0, 3], method=0) - t_last

        # Fill the Gap between Recordings with zeros
        n_zeros = np.zeros(int(vr_fr * t_rel_distance))
        values = dummy.iloc[:, 0].to_numpy()
        vr_values = np.append(vr_values, n_zeros)
        vr_values = np.append(vr_values, values)

        # store last time point of this recording for the next round
        t_last = convert_time_stamps_to_secs(dummy.iloc[-1, 3], method=0)

    # If the first recording file is missing, correct for it now
    if first_file_missing > 0:
        # We know one recording is always 60 seconds long
        # Reset time so that it starts at 60 seconds since the first recording is missing
        n_zeros = np.zeros(int(vr_fr * first_file_missing))
        vr_values = np.append(n_zeros, vr_values)

    # Put all in one Data Frame
    t_max = vr_values.shape[0] / vr_fr
    vr_time = np.arange(0, t_max, 1/vr_fr)
    vr_trace_export = pd.DataFrame(columns=['Time', 'Volt'])
    # Add the time in secs (not the timestamps)
    vr_trace_export['Time'] = vr_time
    vr_trace_export['Volt'] = vr_values

    return vr_trace_export


def transform_ventral_root_parallel(save_dir, base_dir, rec_dur, vr_fr, sw):
    # get file lise
    f_names = os.listdir(f'{base_dir}/{sw}')
    f_names = list(np.sort(f_names))
    # f_names = [s for s in file_list if sw in s]

    f_path = []
    for _, n in enumerate(f_names):
        f_path.append(f'{base_dir}/{sw}/{n}')

    # Check for missing files: Sweep that are missing the first recording must be labeled with the suffix: "fm"
    # We assume that every recording is 60 seconds long
    firs_rec_missing = 0
    if sw.endswith('fm'):
        sw = sw[:-3]
        firs_rec_missing = rec_dur  # secs
        print(f'First Recording missing in sweep {sw}')
        print(f'Will correct for that, assuming that each recording has a duration of 60 seconds!')

    print(f'START PROCESSING: {sw}')
    vr_trace = transform_ventral_root_recording(f_path, vr_fr=vr_fr, first_file_missing=firs_rec_missing)
    if save_dir is not None:
        to_dir = f'{save_dir}/{sw}_ventral_root.csv'
        vr_trace['Volt'].to_csv(to_dir, index=False)  # exclude the time axis
        # vr_trace.to_csv(to_dir, index=False)
        print(f'Ventral Root of Sweep: {sw} stored to HDD')
    # vr_envelope = compute_envelope_of_ventral_root(vr_trace)
    result = {sw: vr_trace}
    return result





import pyqtgraph as pg
import numpy as np


class PyqtgraphSettings:
    # Global pyqtgraph settings
    pg.setConfigOption('background', pg.mkColor('w'))
    pg.setConfigOption('foreground', pg.mkColor('k'))
    pg.setConfigOption('useOpenGL', False)
    pg.setConfigOption('antialias', False)
    pg.setConfigOption('useNumba', False)
    pg.setConfigOption('enableExperimental', False)
    pg.setConfigOption('leftButtonPan', True)
    pg.setConfigOption('imageAxisOrder', 'row-major')


class DataPlotter:
    def __init__(self, master_plot):
        self.master_plot = master_plot

    def clear_plot_data(self, name):
        # check if there is already roi data plotted and remove it
        item_list = self.master_plot.items.copy()
        for item in item_list:
            if item.name() is not None:
                if item.name().startswith(name):
                    self.master_plot.removeItem(item)

    def update_video_plot(self, time_point, y_range):
        self.clear_plot_data(name='video')
        plot_data_item = pg.PlotDataItem(
            [time_point, time_point], [y_range[0], y_range[1]],
            pen=pg.mkPen(color=(255, 0, 255)),
            name=f'video_time',
            skipFiniteCheck=True,
            tip=None,
        )
        # Add plot item to the plot widget
        self.master_plot.addItem(plot_data_item)

    def update(self, time_axis, data, meta_data=None):
        # check if there is already roi data plotted and remove
        self.clear_plot_data(name='data')

        cc = 0
        for t, y in zip(time_axis, data):
            # get meta data
            if meta_data is not None:
                color = meta_data[cc]['color']
                lw = meta_data[cc]['lw']
                plot_name = f'data_{meta_data[cc]["name"]}'
            else:
                color = '#000000'  # black
                lw = 1
                plot_name = f'data_{cc}'
            # Create new  plot item
            plot_data_item = pg.PlotDataItem(
                t, y,
                pen=pg.mkPen(color=color, width=lw),
                # name=f'data_{cc}',
                name=plot_name,
                skipFiniteCheck=True,
                tip=None,
            )
            # plot_data_item.setDownsampling(auto=True, method='peak')
            # Add plot item to the plot widget
            self.master_plot.addItem(plot_data_item)
            cc += 1

    def update_global(self, time_axis, data, meta_data=None):
        # self.clear_global_data()
        self.clear_plot_data(name='global')
        cc = 0  # index of data set (0 = the first global data set, and so on)
        for t, y_data in zip(time_axis, data):
            # Each column in global data set can be a trace
            for y in y_data.T:
                # get meta data
                if meta_data is not None:
                    color = meta_data[cc]['color']
                    lw = meta_data[cc]['lw']
                    plot_name = f'global_{meta_data[cc]["name"]}'
                else:
                    color = '#000000'  # black
                    lw = 1
                    plot_name = f'global_{cc}'

                # Create new  plot item
                plot_data_item = pg.PlotDataItem(
                    t, y,
                    pen=pg.mkPen(color=color, width=lw),
                    # name=f'global_{cc}',
                    name=plot_name,
                    skipFiniteCheck=True,
                    tip=None,
                )

                # Add plot item to the plot widget
                self.master_plot.addItem(plot_data_item)
            cc += 1


from roibaview.plugins.base import BasePlugin
from roibaview.gui_utils import SimpleInputDialog
from PyQt6.QtWidgets import QMessageBox, QLabel
from PyQt6.QtCore import Qt, QObject
import numpy as np
import pyqtgraph as pg


class CutRegionPlugin(BasePlugin):
    name = "Cut ROI Region"
    category = "tool"
    shortcut = "Ctrl+Shift+X"

    def __init__(self, config=None, parent=None):
        self.config = config
        self.parent = parent

    def apply(self, *_):
        controller = getattr(self.parent, "controller", None)
        if not controller or not controller.selected_data_sets:
            QMessageBox.warning(self.parent, "Cut Region", "No data sets selected.")
            return

        # Launch tool
        self.tool = CutRegionTool(controller=controller, parent=self.parent)
        self.tool.start()


class CutRegionTool(QObject):
    def __init__(self, controller, parent):
        super().__init__(parent)
        self.controller = controller
        self.parent = parent
        self.region = None
        self.label = None

    def start(self):
        roi = self.controller.current_roi_idx
        self.selected = list(zip(self.controller.selected_data_sets, self.controller.selected_data_sets_type))

        # Estimate time range
        durations = []
        for name, dtype in self.selected:
            data = self.controller.data_handler.get_data_set(dtype, name)
            meta = self.controller.data_handler.get_data_set_meta_data(dtype, name)
            durations.append(data.shape[0] / meta["sampling_rate"])
        max_duration = max(durations)
        t_vis = np.linspace(0, max_duration, 1000)

        self.region = pg.LinearRegionItem([t_vis[100], t_vis[200]], movable=True, brush=(255, 0, 0, 50))
        self.region.setZValue(10)
        self.controller.data_plotter.master_plot.addItem(self.region)

        # Label
        self.label = QLabel("Cut Region: Adjust with mouse, press ⏎ to confirm, Esc to cancel", self.parent)
        self.label.setStyleSheet("""
            QLabel {
                background-color: rgba(0, 0, 0, 180);
                color: white;
                padding: 4px 8px;
                border-radius: 6px;
                font-size: 10pt;
            }
        """)
        self.label.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        self.label.adjustSize()
        self.label.move(120, 100)
        self.label.show()

        self.parent.key_pressed.connect(self._on_key_press)

    def _on_key_press(self, event):
        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            self._finalize_cut()
        elif event.key() == Qt.Key.Key_Escape:
            self._cleanup()

    def _finalize_cut(self):
        try:
            min_x, max_x = self.region.getRegion()
            if max_x <= min_x:
                QMessageBox.warning(self.parent, "Cut Region", "Invalid region selected.")
                self._cleanup()
                return

            # Ask for suffix
            dialog = SimpleInputDialog("Cut Region", "Suffix for new dataset name:", default_value="cut")
            if dialog.exec() != dialog.DialogCode.Accepted:
                self._cleanup()
                return

            suffix = dialog.get_input().strip()
            if not suffix:
                QMessageBox.warning(self.parent, "Cut Region", "Suffix cannot be empty.")
                self._cleanup()
                return

            for name, dtype in self.selected:
                data = self.controller.data_handler.get_data_set(dtype, name)
                meta = self.controller.data_handler.get_data_set_meta_data(dtype, name)
                fr = meta["sampling_rate"]

                start = int(min_x * fr)
                end = int(max_x * fr)
                if end <= start:
                    continue

                cut_data = data[start:end] if data.ndim == 1 else data[start:end, :]
                new_name = f"{name}_{suffix}"

                self.controller.data_handler.add_new_data_set(
                    data_set_type=dtype,
                    data_set_name=new_name,
                    data=cut_data,
                    sampling_rate=fr,
                    time_offset=0,
                    y_offset=0,
                    header=meta.get("roi_names", list(range(cut_data.shape[1] if cut_data.ndim > 1 else 1)))
                )
                self.controller.add_data_set_to_list(dtype, new_name)
        finally:
            self._cleanup()

    def _cleanup(self):
        if self.region:
            self.controller.data_plotter.master_plot.removeItem(self.region)
        if self.label:
            self.label.deleteLater()
        try:
            self.parent.key_pressed.disconnect(self._on_key_press)
        except Exception:
            pass

    # def __del__(self):
    #     print("CutRegionTool deleted.")



try:
    import caiman as cm
    from caiman.motion_correction import MotionCorrect
    from caiman.utils.visualization import inspect_correlation_pnr
    HAS_CAIMAN = True
except ImportError:
    HAS_CAIMAN = False

from roibaview.plugins.base import BasePlugin
from PyQt6.QtWidgets import QFileDialog, QMessageBox
import os


class MotionCorrectionPlugin(BasePlugin):
    name = "Motion Correction (CaImAn)"
    category = "tool"
    unavailable_reason = "CaImAn is not installed. Install it via pip to enable this plugin."

    def __init__(self, config=None, parent=None):
        self.config = config
        self.parent = parent

    @classmethod
    def available(cls):
        return HAS_CAIMAN

    def apply(self, *_):
        file_path, _ = QFileDialog.getOpenFileName(
            self.parent,
            "Select TIFF file for motion correction",
            "",
            "TIF files (*.tif *.tiff)"
        )

        if not file_path:
            return  # user canceled

        try:
            # Output folder next to original file
            base_dir = os.path.dirname(file_path)
            base_name = os.path.basename(file_path).split('.')[0]
            out_dir = os.path.join(base_dir, base_name + "_mc")
            os.makedirs(out_dir, exist_ok=True)

            # Set up CaImAn motion correction
            mc = MotionCorrect([file_path], dview=None,
                               pw_rigid=True,
                               max_shifts=(6, 6),
                               strides=(48, 48),
                               overlaps=(24, 24),
                               max_deviation_rigid=3,
                               shifts_opencv=True,
                               border_nan='copy')

            mc.motion_correct(save_movie=True)

            corrected_path = mc.fname_tot_els if mc.fname_tot_els else file_path
            QMessageBox.information(
                self.parent,
                "Motion Correction Complete",
                f"Corrected video saved to: {corrected_path}"
            )

        except Exception as e:
            QMessageBox.critical(self.parent, "Motion Correction Error", str(e))



# plugins/base.py

class BasePlugin:
    """
    Abstract base class for all RoiBaView plugins.
    """
    name = "Unnamed Plugin"
    category = "filter"  # 'filter', 'tool', 'analysis', etc.
    unavailable_reason = "This plugin is unavailable."

    @classmethod
    def available(cls):
        """
        Optional method to determine if the plugin should be loaded.
        Override in subclasses if conditional loading is needed.
        """
        return True

    def apply(self, data=None, sampling_rate=None):
        """
        Apply the plugin logic.

        For 'filter' plugins, both parameters are used.
        For 'tool' plugins, these can be ignored.

        Parameters:
            data (np.ndarray): ROI data (optional for tools)
            sampling_rate (float): Sampling rate (Hz)

        Returns:
            np.ndarray or None: Processed data or nothing
        """
        raise NotImplementedError("Plugin must implement the 'apply' method.")


from PyQt6.QtWidgets import (
    QDialog, QVBoxLayout, QLabel, QLineEdit, QPushButton, QMessageBox,
    QDialogButtonBox, QFileDialog, QHBoxLayout, QCheckBox, QWidget, QColorDialog
)
from PyQt6.QtGui import QIntValidator, QDoubleValidator
import os


class DynamicInputDialog(QDialog):
    def __init__(self, title='Input Dialog', fields=None, parent=None):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.fields = fields or {}
        self.inputs = {}  # key -> (widget, type_str)

        layout = QVBoxLayout()

        for label, (default, type_str) in self.fields.items():
            if type_str == 'bool':
                # Checkbox for boolean input
                checkbox = QCheckBox(label)
                checkbox.setChecked(bool(default))
                layout.addWidget(checkbox)
                self.inputs[label] = (checkbox, type_str)

            elif type_str in ('file', 'dir'):
                # File or directory picker
                layout.addWidget(QLabel(label))
                container = QHBoxLayout()
                line_edit = QLineEdit(str(default))
                button = QPushButton("Browse...")
                container.addWidget(line_edit)
                container.addWidget(button)
                layout.addLayout(container)

                def open_dialog(_, le=line_edit, t=type_str):
                    if t == 'file':
                        file_path, _ = QFileDialog.getOpenFileName(self, "Select File")
                        if file_path:
                            le.setText(file_path)
                    else:
                        dir_path, _ = QFileDialog.getSaveFileName(self, "Select Directory", filter="CSV Files (*.csv)")
                        if dir_path:
                            if not dir_path.endswith('.csv'):
                                dir_path += '.csv'
                            le.setText(dir_path)

                button.clicked.connect(open_dialog)
                self.inputs[label] = (line_edit, type_str)

            else:
                # Regular text/numeric input
                layout.addWidget(QLabel(label))
                line_edit = QLineEdit(str(default))
                if type_str == 'int':
                    line_edit.setValidator(QIntValidator())
                elif type_str == 'float':
                    line_edit.setValidator(QDoubleValidator())
                layout.addWidget(line_edit)
                self.inputs[label] = (line_edit, type_str)

        self.add_ok_cancel_buttons(layout)
        self.setLayout(layout)

    def add_ok_cancel_buttons(self, layout):
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        buttonBox = QDialogButtonBox(QBtn)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        layout.addWidget(buttonBox)

    def get_inputs(self):
        result = {}
        for label, (widget, type_str) in self.inputs.items():
            if type_str == 'bool':
                result[label] = widget.isChecked()
            else:
                value = widget.text()
                try:
                    if type_str == 'int':
                        result[label] = int(value)
                    elif type_str == 'float':
                        result[label] = float(value)
                    else:
                        result[label] = value
                except ValueError:
                    result[label] = None
        return result


class SimpleInputDialog(QDialog):
    """
    That's how you call it:
        dialog = SimpleInputDialog(title='Settings', text='Please enter some stuff: ')
        if dialog.exec() == QDialog.DialogCode.Accepted:
            received = dialog.get_input()
        else:
            return None
    """
    def __init__(self, title, text, default_value=0,  parent=None):
        super().__init__(parent)
        self.title = title
        self.text = text
        self.default_value = default_value

        self.setWindowTitle(self.title)
        layout = QVBoxLayout()

        # Create input fields
        self.user_input = QLineEdit()
        self.user_input.setText(str(self.default_value))

        # Add labels
        layout.addWidget(QLabel(self.text))
        layout.addWidget(self.user_input)

        # Add OK and Cancel buttons
        self.add_ok_cancel_buttons(layout)
        self.setLayout(layout)

    def get_input(self):
        # Return the entered settings
        output = self.user_input.text()
        return output

    def add_ok_cancel_buttons(self, layout):
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        buttonBox = QDialogButtonBox(QBtn)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        layout.addWidget(buttonBox)


class BrowseFileDialog(QFileDialog):

    def __init__(self, main_gui):
        QFileDialog.__init__(self)
        self.master = main_gui
        self.default_dir = 'C:/'
        # self.file_format = 'csv file, (*.csv)'

    def browse_file(self, file_format):
        file_dir = self.getOpenFileName(self.master, 'Open File', self.default_dir, file_format)[0]
        # change default dir to current dir
        self.default_dir = os.path.split(file_dir)[0]
        return file_dir

    def save_file_name(self, file_format):
        file_dir = self.getSaveFileName(self.master, 'Open File', self.default_dir, file_format)[0]
        # change default dir to current dir
        self.default_dir = os.path.split(file_dir)[0]
        return file_dir

    def browse_directory(self):
        return self.getExistingDirectory()


class InputDialog(QDialog):
    def __init__(self, dialog_type,  parent=None):
        super().__init__(parent)
        self.dialog_type = dialog_type
        self.fields = dict()
        if self.dialog_type == 'import_csv':
            self._set_gui_import_csv_dialog()
        elif self.dialog_type == 'rename':
            self._set_gui_rename_data_set()
        elif self.dialog_type == 'stimulus':
            self._set_gui_stimulus_dialog()

    def get_input(self):
        # Return the entered settings
        output = dict()
        for k in self.fields:
            if k == 'is_global':
                output[k] = self.fields[k].isChecked()
            else:
                output[k] = self.fields[k].text()
        return output

    def _set_gui_stimulus_dialog(self):
        self.setWindowTitle("Settings")

        layout = QVBoxLayout()

        # Create input fields
        self.fields['name'] = QLineEdit()

        # Add labels
        layout.addWidget(QLabel("Name: "))
        layout.addWidget(self.fields['name'])

        # Add OK and Cancel buttons
        self.add_ok_cancel_buttons(layout)
        self.setLayout(layout)

    def _set_gui_import_csv_dialog(self):
        self.setWindowTitle("Settings")

        layout = QVBoxLayout()

        # Create input fields
        self.fields['data_set_name'] = QLineEdit()
        self.fields['fr'] = QLineEdit()
        self.fields['is_global'] = QCheckBox()
        self.fields['select_column'] = QLineEdit()

        # Add labels
        layout.addWidget(QLabel("Data Name:"))
        layout.addWidget(self.fields['data_set_name'])
        layout.addWidget(QLabel("Sampling Rate:"))
        layout.addWidget(self.fields['fr'])
        layout.addWidget(QLabel("Global Data Set:"))
        layout.addWidget(self.fields['is_global'])
        layout.addWidget(QLabel("Select single column only:"))
        layout.addWidget(self.fields['select_column'])

        # Add OK and Cancel buttons
        self.add_ok_cancel_buttons(layout)
        self.setLayout(layout)

    def _set_gui_rename_data_set(self):
        self.setWindowTitle("Rename Data Set")

        layout = QVBoxLayout()

        # Create input fields
        self.fields['data_set_name'] = QLineEdit()

        # Add labels
        layout.addWidget(QLabel("Data Name:"))
        layout.addWidget(self.fields['data_set_name'])

        # Add OK and Cancel buttons
        self.add_ok_cancel_buttons(layout)
        self.setLayout(layout)

    def add_ok_cancel_buttons(self, layout):
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        buttonBox = QDialogButtonBox(QBtn)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        layout.addWidget(buttonBox)


class ChangeStyle(QWidget):
    def __init__(self):
        super().__init__()

    def get_color(self):
        color = QColorDialog.getColor()
        if color.isValid():
            # print(f"Selected color: {color.name()}")
            pass
        return color.name()

    def get_lw(self):
        dialog = SimpleInputDialog(title='Settings', text='Line Width: ')
        if dialog.exec() == QDialog.DialogCode.Accepted:
            return float(dialog.get_input())
        else:
            return None


class MessageBox:
    def __init__(self, title, text):
        dlg = QMessageBox()
        dlg.setWindowTitle(title)
        dlg.setText(text)
        dlg.setStandardButtons(QMessageBox.StandardButton.Ok)
        button = dlg.exec()
        if button == QMessageBox.StandardButton.Ok:
            return




import sys
from PyQt6.QtWidgets import QApplication
from roibaview.gui import MainWindow
from roibaview.controller import Controller


def main():
    print("RoiBaView is starting...")
    # Start Qt Application
    app = QApplication(sys.argv)
    screen = app.primaryScreen().availableGeometry()

    # GUI
    window = MainWindow(screen)

    # Start Controller
    Controller(gui=window)
    window.show()
    app.exec()


if __name__ == '__main__':
    main()



import os
import numpy as np
import pandas as pd
import configparser
from PyQt6.QtWidgets import QMessageBox, QListWidgetItem, QDialog
from PyQt6.QtCore import pyqtSignal, QObject, Qt
from roibaview.data_handler import DataHandler
from roibaview.gui_utils import SimpleInputDialog, BrowseFileDialog, InputDialog, ChangeStyle, MessageBox
from roibaview.data_plotter import DataPlotter, PyqtgraphSettings
from roibaview.video_viewer import VideoViewer
from roibaview.plugins.loader import load_plugins


class Controller(QObject):
    # ==================================================================================================================
    # SIGNALS
    # ------------------------------------------------------------------------------------------------------------------
    signal_import_traces = pyqtSignal()
    signal_roi_idx_changed = pyqtSignal()
    signal_closing = pyqtSignal()

    # ==================================================================================================================
    # INITIALIZING
    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, gui):
        QObject.__init__(self)
        # Create GUI
        self.gui = gui
        self.gui.controller = self  # Expose controller to plugins via GUI
        self.gui.closeEvent = self.closeEvent
        self.mouse_x_pos = 0

        # Get a DataHandler
        self.data_handler = DataHandler()
        self.selected_data_sets = []
        self.selected_data_sets_type = []
        self.selected_data_sets_rows = []
        self.selected_data_sets_items = []
        self.current_roi_idx = 0

        # View Box Right Click Context Menu
        # Hide "Plot Options"
        self.gui.trace_plot_item.ctrlMenu.menuAction().setVisible(False)

        # Get a Video Viewer
        self.video_viewer = VideoViewer()
        self.video_viewers = []

        # Get DataPlotter
        self.data_plotter = DataPlotter(self.gui.trace_plot_item)

        # Get a File Browser
        self.file_browser = BrowseFileDialog(self.gui)

        # Get a VR Detector
        self.vr_detection = None

        # Set Selection Mode
        self.signal_selection_status = False
        self.cut_out_region = None

        # Establish Connections to Buttons and Menus
        self.connections()

        # KeyBoard Bindings
        self.gui.trace_plot_item.scene().sigMouseMoved.connect(self.mouse_moved)

        self.pyqtgraph_settings = PyqtgraphSettings()

        # Check Config File
        package_dir = os.path.dirname(__file__)
        config_path = os.path.join(package_dir, 'config.ini')
        self.config_name = config_path

        if not os.path.exists(self.config_name):
            self._create_config_file()
        else:
            self.config = configparser.ConfigParser()
            self.config.read(self.config_name)

        # Load plugins with context (pass config and GUI as needed)
        self.plugins = load_plugins(context={
            'config': self.config,  # or other settings object
            'parent': self.gui  # used for dialog-based plugins
        })

        # Categorize them for use
        self.tool_plugins = [p for p in self.plugins if p.category == "tool"]
        self.utils_plugins = [p for p in self.plugins if p.category == "utils"]

        # Context Menu
        self.filter_plugins = [p for p in self.plugins if p.category == "filter"]
        self.transformation_plugins = [p for p in self.plugins if p.category == "transformation"]

        # Register tools in Tools menu
        self.gui.add_tools_menu_plugins(self.tool_plugins)

        # Register utils in Utils menu
        self.gui.add_utils_menu_plugins(self.utils_plugins)

        # Populate Context Menu Filter Plugins
        self.gui.populate_filter_plugins_menu(self.filter_plugins, self.apply_filter_plugin)
        self.gui.populate_transformation_plugins_menu(self.transformation_plugins, self.apply_filter_plugin)

    def apply_filter_plugin(self, plugin):
        if not self.selected_data_sets:
            return

        for name, kind in zip(self.selected_data_sets, self.selected_data_sets_type):
            data = self.data_handler.get_data_set(kind, name)
            meta = self.data_handler.get_data_set_meta_data(kind, name)
            result = plugin.apply(data, meta['sampling_rate'])
            if type(result) is tuple:  # function returns two values, the second is the new sampling rate (down sampling)
                fr = result[1]
                result = result[0]
            else:
                fr = meta['sampling_rate']
            new_name = f"{name}_{plugin.name.replace(' ', '_')}"
            already_exists = self.data_handler.add_new_data_set(
                data_set_type=kind,
                data_set_name=new_name,
                data=result,
                sampling_rate=fr,
                time_offset=0,
                y_offset=0,
                header=meta.get("roi_names", list(range(result.shape[1])))
            )
            if already_exists:
                data_set_name = new_name + '_new'
                self.add_data_set_to_list(kind, data_set_name)
            else:
                self.add_data_set_to_list(kind, new_name)

    def _create_config_file(self):
        self.config = configparser.ConfigParser()
        self.config['FFMPEG'] = {'dir': 'NaN'}
        with open(self.config_name, 'w') as configfile:
            self.config.write(configfile)

    def connections(self):
        # File Menu
        self.gui.file_menu_import_csv.triggered.connect(self.import_csv_file)
        self.gui.file_menu_new_viewer_file.triggered.connect(self.new_file)
        self.gui.file_menu_save_viewer_file.triggered.connect(self.save_file)
        self.gui.file_men_open_viewer_file.triggered.connect(self.open_file)
        self.gui.file_menu_action_exit.triggered.connect(self.gui.close)

        # DataSets List
        # Connect item selection changed signal
        self.gui.data_sets_list.itemSelectionChanged.connect(self.data_set_selection_changed)

        # Arrow Buttons
        self.gui.next_button.clicked.connect(self.next_roi)
        self.gui.prev_button.clicked.connect(self.prev_roi)

        # ROI changed
        self.signal_roi_idx_changed.connect(lambda: self.update_plots(change_global=False))

        # Context Menu
        self.gui.data_sets_list_rename.triggered.connect(self.rename_data_set)
        self.gui.data_sets_list_delete.triggered.connect(self.delete_data_set)
        self.gui.data_sets_list_delete_col.triggered.connect(self.delete_column)

        self.gui.data_sets_list_export.triggered.connect(self.export_to_csv)
        self.gui.data_sets_list_time_offset.triggered.connect(self.time_offset)
        self.gui.data_sets_list_y_offset.triggered.connect(self.y_offset)

        # Style Submenu
        self.gui.style_color.triggered.connect(self.pick_color)
        self.gui.style_lw.triggered.connect(self.pick_lw)

        # Tools
        # Video Viewer
        self.gui.tools_menu_open_video_viewer.triggered.connect(self.open_video_viewer)
        self.video_viewer.TimePoint.connect(self.connect_video_to_plot)

        # Convert Ventral Root
        self.gui.tools_menu_convert_ventral_root.triggered.connect(self.convert_ventral_root)

        # Create Stimulus from File
        self.gui.tools_menu_create_stimulus.triggered.connect(self.create_stimulus_from_file)

        # KeyBoard Bindings
        self.gui.key_pressed.connect(self.on_key_press)
        self.data_plotter.master_plot.scene().sigMouseClicked.connect(self.on_mouse_click)

    def delete_column(self):
        if len(self.selected_data_sets) > 1:
            dlg = QMessageBox()
            dlg.setWindowTitle('ERROR')
            dlg.setText(f'You cannot rename multiple data sets at once!')
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Ok:
                return None
        if len(self.selected_data_sets) == 0:
            return None

        # Get Column Nr
        dialog = SimpleInputDialog(title='Settings', text='Please enter some stuff: ')
        if dialog.exec() == QDialog.DialogCode.Accepted:
            received = dialog.get_input()
        else:
            return None

        col_nr = int(received)
        data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)

        # Remove the column
        self.data_handler.delete_column(data_set_type, data_set_name, col_nr)

    def export_to_csv(self):
        file_dir = self.file_browser.save_file_name('csv file, (*.csv *.txt)')
        if file_dir:
            if len(self.selected_data_sets) > 1:
                dlg = QMessageBox()
                dlg.setWindowTitle('ERROR')
                dlg.setText(f'You cannot export multiple data sets at once!')
                button = dlg.exec()
                if button == QMessageBox.StandardButton.Ok:
                    return None
            if len(self.selected_data_sets) == 0:
                return None

            # Export selected data set to csv
            data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)
            df = pd.DataFrame(self.data_handler.get_data_set(data_set_type=data_set_type, data_set_name=data_set_name))
            df.to_csv(file_dir, index=False, header=False)

    def create_stimulus_from_file(self):
        file_dir = self.file_browser.browse_file('csv file, (*.csv *.txt)')
        if file_dir:
            dialog = InputDialog(dialog_type='stimulus')
            if dialog.exec() == QDialog.DialogCode.Accepted:
                # name
                received = dialog.get_input()
                data_set_name = received['name']
            else:
                return None

            protocol = pd.read_csv(file_dir)
            t_max = max(protocol.max())
            # t0 = 0
            fr = 100
            stimulus = np.zeros(int(fr * t_max))
            for on, off in zip(protocol.iloc[:, 0], protocol.iloc[:, 1]):
                start = int(on * fr)
                end = int(off * fr)
                stimulus[start:end] = 1

            # Prepare it for hdf5 matrix form
            stimulus = list(stimulus[:, np.newaxis])
            data_set_type = 'global_data_sets'
            # Add stimulus to data sets
            self.data_handler.add_new_data_set(
                data_set_type=data_set_type,
                data_set_name=data_set_name,
                data=stimulus,
                sampling_rate=fr,
                time_offset=0,
                y_offset=0,
                header='Stimulus')

            # Add new data set to the list in the GUI
            self.add_data_set_to_list(data_set_type, data_set_name)

    def convert_ventral_root(self):
        """
        Expects following file structure:
        └── vr_data
            ├── sweep_01
            │   ├── sw_01_01.txt
            │   ├── sw_01_02.txt
            │   :
            │   └── sw_01_n.txt
            ├── sweep_02
            :
            └── sweep_n

        All vr text files from one sweep will then be combined into one meaningful and solid data file (csv)
        :return:
        """
        from joblib import Parallel, delayed
        from roibaview.ventral_root import transform_ventral_root_parallel, pickle_stuff
        file_structure = '''
        Expects following file structure:
        └── vr_data
            ├── sweep_01
            │   ├── sw_01_01.txt
            │   ├── sw_01_02.txt
            │   :
            │   └── sw_01_n.txt
            ├── sweep_02
            :
            └── sweep_n
            
        All vr text files from one sweep will then be combined into one meaningful and solid data file (csv)
        '''

        store_to_dict = False
        vr_rec_dur = 60  # in secs
        vr_fr = 10000  # in Hz

        # Get directory
        file_dir = self.file_browser.browse_directory()
        save_dir = file_dir
        if file_dir:
            print('========== VENTRAL ROOT ==========')
            print('')
            print(file_structure)
            print('')
            print(f'Ventral Root Recording Files in: {file_dir}')
            print(f'Store result to: {save_dir}')
            print('++++ START PROCESSING +++')
            print('This can take some time ...')
            print( 'This relies heavily on CPU, RAM and HDD. HDD is normally the bottleneck, so make sure to use a fast one!')
            print('... Please Wait ...')
            print('')
            # Process all sweeps in parallel
            sweep_numbers = os.listdir(file_dir)
            results = Parallel(n_jobs=-2)(delayed(
                transform_ventral_root_parallel)(save_dir, file_dir, vr_rec_dur, vr_fr, i) for i in
                                          sweep_numbers)
            if store_to_dict:
                print('STORING DICT TO HDD')
                vr_recordings = dict()
                for res in results:
                    sw = list(res.keys())[0]
                    vr_recordings[sw] = res[sw]
                pickle_stuff(f'{save_dir}/all_ventral_root.pickle', data=vr_recordings)

            print('++++ FINISHED PROCESSING ++++')

    def pick_lw(self):
        if len(self.selected_data_sets) > 1:
            dlg = QMessageBox()
            dlg.setWindowTitle('ERROR')
            dlg.setText(f'You cannot change color of multiple data sets at once!')
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Ok:
                return None

        if len(self.selected_data_sets) == 0:
            return None

        lw = ChangeStyle().get_lw()
        data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)
        self.data_handler.add_meta_data(data_set_type=data_set_type, data_set_name=data_set_name, metadata_dict={'lw': lw})
        self.update_plots(change_global=True)

    def pick_color(self):
        if len(self.selected_data_sets) > 1:
            dlg = QMessageBox()
            dlg.setWindowTitle('ERROR')
            dlg.setText(f'You cannot change color of multiple data sets at once!')
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Ok:
                return None

        if len(self.selected_data_sets) == 0:
            return None

        color = ChangeStyle().get_color()
        data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)
        self.data_handler.add_meta_data(data_set_type=data_set_type, data_set_name=data_set_name, metadata_dict={'color': color})
        self.update_plots(change_global=True)

    def connect_video_to_plot(self, time_point):
        for v in self.video_viewers:
            if v.connected_to_data_trace:
                if len(self.selected_data_sets) > 0:
                    y_min = []
                    y_max = []
                    for data_set_name, data_set_type in zip(self.selected_data_sets, self.selected_data_sets_type):
                        if data_set_type == 'data_sets':
                            data = self.data_handler.get_roi_data(data_set_name, roi_idx=self.current_roi_idx)
                        else:
                            data = self.data_handler.get_data_set('global_data_sets', data_set_name)
                        y_min.append(np.min(data))
                        y_max.append(np.max(data))
                    y_range = [np.min(y_min), np.max(y_max)]
                    self.data_plotter.update_video_plot(time_point, y_range)
                else:
                    self.data_plotter.clear_plot_data(name='video')
            else:
                self.data_plotter.clear_plot_data(name='video')

    def open_video_viewer(self):
        self.video_viewers.append(VideoViewer())
        self.video_viewers[-1].show()
        self.video_viewers[-1].TimePoint.connect(self.connect_video_to_plot)

    def y_offset(self):
        if len(self.selected_data_sets) > 0:
            for k, _ in enumerate(self.selected_data_sets):
                data_set_name = self.selected_data_sets[k]
                data_set_type = self.selected_data_sets_type[k]
                meta_data = self.data_handler.get_data_set_meta_data(
                    data_set_type=data_set_type,
                    data_set_name=data_set_name
                )

                # Get settings by user input
                dialog = SimpleInputDialog('Settings', 'Please enter Y Offset: ', default_value=meta_data['y_offset'])
                if dialog.exec() == dialog.DialogCode.Accepted:
                    y_offset = {'y_offset': float(dialog.get_input())}
                    self.data_handler.add_meta_data(data_set_type, data_set_name, y_offset)
                    self.update_plots(change_global=True)
                else:
                    return None

    def time_offset(self):
        if len(self.selected_data_sets) > 0:
            for k, _ in enumerate(self.selected_data_sets):
                data_set_name = self.selected_data_sets[k]
                data_set_type = self.selected_data_sets_type[k]
                meta_data = self.data_handler.get_data_set_meta_data(
                    data_set_type=data_set_type,
                    data_set_name=data_set_name
                )

                # Get settings by user input
                dialog = SimpleInputDialog('Settings', 'Please enter Time Offset [s]: ', default_value=meta_data['time_offset'])
                if dialog.exec() == dialog.DialogCode.Accepted:
                    time_offset = {'time_offset': float(dialog.get_input())}
                    self.data_handler.add_meta_data(data_set_type, data_set_name, time_offset)
                    self.update_plots(change_global=True)
                else:
                    return None

    def rename_data_set(self):
        if len(self.selected_data_sets) > 1:
            dlg = QMessageBox()
            dlg.setWindowTitle('ERROR')
            dlg.setText(f'You cannot rename multiple data sets at once!')
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Ok:
                return None
        if len(self.selected_data_sets) == 0:
            return None

        data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)
        # Get settings by user input
        dialog = InputDialog(dialog_type='rename')
        if dialog.exec() == QDialog.DialogCode.Accepted:
            received = dialog.get_input()
            new_name = received['data_set_name']
            if new_name != '':
                self.data_handler.rename_data_set(data_set_type=data_set_type, data_set_name=data_set_name, new_name=new_name)
                self.rename_item_from_list(data_set_item=data_set_item, new_name=new_name)
            else:
                dlg = QMessageBox()
                dlg.setWindowTitle('ERROR')
                dlg.setText(f'Please enter a valid name!')
                button = dlg.exec()
                if button == QMessageBox.StandardButton.Ok:
                    return None
        else:
            return None

    def delete_data_set(self):
        if len(self.selected_data_sets) > 0:
            dlg = QMessageBox(self.gui)
            dlg.setWindowTitle('Delete Data Set')
            dlg.setText(f'Are You Sure You Want To Delete The Selected Data Sets" ?')
            dlg.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
            dlg.setIcon(QMessageBox.Icon.Question)
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Yes:
                for dt, ds, item in zip(self.selected_data_sets_type, self.selected_data_sets, self.selected_data_sets_items):
                    self.data_handler.delete_data_set(dt, ds)
                    self.remove_selected_data_set_from_list(ds, item)

    def get_selected_data_sets(self, k):
        data_set_name = self.selected_data_sets[k]
        data_set_type = self.selected_data_sets_type[k]
        data_set_item = self.selected_data_sets_items[k]
        return data_set_name, data_set_type, data_set_item

    def context_menu(self, mode):
        result = None
        if len(self.selected_data_sets) > 0:
            for k, _ in enumerate(self.selected_data_sets):
                data_set_name = self.selected_data_sets[k]
                data_set_type = self.selected_data_sets_type[k]
                data = self.data_handler.get_data_set(data_set_type=data_set_type, data_set_name=data_set_name)
                meta_data = self.data_handler.get_data_set_meta_data(data_set_type=data_set_type, data_set_name=data_set_name)
                if mode == 'df_f':
                    # Get settings by user input
                    dialog = InputDialog(dialog_type='df_over_f')
                    if dialog.exec() == QDialog.DialogCode.Accepted:
                        received = dialog.get_input()
                        fbs_per = float(received['fbs_per'])
                        fbs_win = float(received['fbs_window'])
                    else:
                        return None

                    result = self.data_transformer.to_delta_f_over_f(data, fbs_per=fbs_per, fr=meta_data['sampling_rate'], window=fbs_win)

                if mode == 'z':
                    result = self.data_transformer.to_z_score(data)

                if mode == 'min_max':
                    result = self.data_transformer.to_min_max(data)

                # Create a new data set from this
                if result is not None:
                    check = self.data_handler.add_new_data_set(
                        data_set_type=data_set_type,
                        data_set_name=f'{data_set_name}_{mode}',
                        data=result,
                        sampling_rate=meta_data['sampling_rate'],
                        time_offset=meta_data['time_offset'],
                        y_offset=meta_data['y_offset'],
                        header=f'{data_set_name}_{mode}'
                    )
                    # Add new data set to the list in the GUI
                    if check:
                        # data set name already exists, so whe have to change it
                        data_set_name = f'{data_set_name}_{mode}' + '_new'
                    else:
                        data_set_name = f'{data_set_name}_{mode}'
                    self.add_data_set_to_list(data_set_type, data_set_name)

    def next_roi(self):
        # First check if there are active data sets
        if 'data_sets' in self.selected_data_sets_type:
            self.current_roi_idx = (self.current_roi_idx + 1) % self.data_handler.roi_count
            self.signal_roi_idx_changed.emit()

    def prev_roi(self):
        # First check if there are active data sets
        if 'data_sets' in self.selected_data_sets_type:
            self.current_roi_idx = (self.current_roi_idx - 1) % self.data_handler.roi_count
            self.signal_roi_idx_changed.emit()

    def update_plots(self, change_global=True):
        # get new roi data
        roi_data = []
        time_points = []
        global_data = []
        global_time_points = []
        meta_data_list = list()
        global_meta_data_list = list()

        for data_set_name, data_set_type in zip(self.selected_data_sets, self.selected_data_sets_type):
            if data_set_type == 'data_sets':
                r = self.data_handler.get_roi_data(data_set_name, roi_idx=self.current_roi_idx)
                meta_data = self.data_handler.get_data_set_meta_data('data_sets', data_set_name)
                fr = meta_data['sampling_rate']
                time_offset = meta_data['time_offset']
                y_offset = meta_data['y_offset']
                try:
                    time_points.append(self.data_handler.compute_time_axis(r.shape[0], fr) + time_offset)
                except AttributeError:
                    return

                roi_data.append(r + y_offset)
                meta_data_list.append(meta_data)
            if data_set_type == 'global_data_sets' and change_global:
                r = self.data_handler.get_data_set('global_data_sets', data_set_name)
                meta_data = self.data_handler.get_data_set_meta_data('global_data_sets', data_set_name)
                fr = meta_data['sampling_rate']
                time_offset = meta_data['time_offset']
                y_offset = meta_data['y_offset']
                global_time_points.append(self.data_handler.compute_time_axis(r.shape[0], fr) + time_offset)
                global_data.append(r + y_offset)
                global_meta_data_list.append(meta_data)

        # Update Plot
        if len(roi_data) > 0:
            self.data_plotter.update(time_points, roi_data, meta_data_list)
            self.data_plotter.master_plot.setTitle(f'ROI: {self.current_roi_idx+1}')
        else:
            self.data_plotter.clear_plot_data(name='data')

        # Update Global Plot
        if change_global:
            if len(global_data) > 0:
                self.data_plotter.update_global(global_time_points, global_data, global_meta_data_list)
            else:
                self.data_plotter.clear_plot_data(name='global')

    def data_set_selection_changed(self):
        # Get selected data sets
        self.selected_data_sets = [item.text() for item in self.gui.sender().selectedItems()]
        self.selected_data_sets_type = [item.data(1) for item in self.gui.sender().selectedItems()]
        self.selected_data_sets_items = [k for k in self.gui.sender().selectedItems()]

        # Update Plots
        self.update_plots()

    def import_csv_file(self):
        # Get file dir
        file_dir = self.file_browser.browse_file('csv file, (*.csv *.txt)')
        if file_dir:
            # Get data set name by user
            dialog = InputDialog(dialog_type='import_csv')
            if dialog.exec() == QDialog.DialogCode.Accepted:
                received = dialog.get_input()
                data_set_name = received['data_set_name']
                fr = float(received['fr'])
                is_global = received['is_global']
                single_column = received['select_column']
            else:
                return None

            if is_global:
                data_set_type = 'global_data_sets'
            else:
                data_set_type = 'data_sets'

            # check if the data set name already exists
            check_if_exists = self.data_handler.check_if_exists(data_set_type, data_set_name)
            if check_if_exists:
                MessageBox(title='ERROR', text='Data set with this name already exists! Will rename it.')
                data_set_name = data_set_name + '_new'

            # Import csv file
            status = self.data_handler.import_csv(
                file_dir=file_dir,
                data_name=data_set_name,
                sampling_rate=fr,
                data_set_type=data_set_type,
                column=single_column
            )

            # Add new data set to the list in the GUI
            if status:
                self.add_data_set_to_list(data_set_type, data_set_name)

    def add_data_set_to_list(self, data_set_type, data_set_name):
        # row = self.gui.data_sets_list.count()
        # Add new data set to the list in the GUI
        new_list_item = QListWidgetItem()
        new_list_item.setText(data_set_name)
        new_list_item.setData(1, data_set_type)
        # new_list_item.setData(3, row)
        self.gui.data_sets_list.addItem(new_list_item)

    def rename_item_from_list(self, data_set_item, new_name):
        item = self.gui.data_sets_list.item(self.gui.data_sets_list.row(data_set_item))
        item.setText(new_name)

    def remove_selected_data_set_from_list(self, data_set_name, data_set_item):
        if not data_set_name:
            return
        else:
            self.gui.data_sets_list.takeItem(self.gui.data_sets_list.row(data_set_item))

    def save_file(self):
        file_dir = self.file_browser.save_file_name('hdf5 file, (*.hdf5)')
        if file_dir:
            self.data_handler.save_file(file_dir)

    def open_file(self):
        file_dir = self.file_browser.browse_file('hdf5 file, (*.hdf5)')
        if file_dir:
            self.data_handler.open_file(file_dir)
            data_structure = self.data_handler.get_info()
            # Add new data set to the list in the GUI
            for data_set_type in data_structure:
                for ds in data_structure[data_set_type]:
                    self.add_data_set_to_list(data_set_type, ds)

            # Get the ROI count (from the first data set)
            if 'data_sets' in data_structure:
                ds = data_structure['data_sets'][0]
                self.data_handler.roi_count = self.data_handler.get_roi_count(ds)

    def new_file(self):
        dlg = QMessageBox(self.gui)
        dlg.setWindowTitle('New Session')
        dlg.setText(f'Are you sure you want start a new session. All unsaved data will be lost!')
        dlg.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
        dlg.setIcon(QMessageBox.Icon.Question)
        button = dlg.exec()
        if button == QMessageBox.StandardButton.Yes:
            self.data_handler.new_file()
            self.gui.data_sets_list.clear()

    def _create_short_cuts(self):
        pass

    def _connect_short_cuts(self, connect=True):
        pass

    # ==================================================================================================================
    # MOUSE AND KEY PRESS HANDLING
    # ------------------------------------------------------------------------------------------------------------------
    def on_mouse_click(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            pos = event.scenePos()
            mouse_point = self.data_plotter.master_plot.vb.mapSceneToView(pos)
            # Get x value (corresponding to time axis)
            self.mouse_x_pos = mouse_point.x()

    def on_key_press(self, event):
        if event.key() == Qt.Key.Key_Left:
            # left arrow key
            self.prev_roi()
        elif event.key() == Qt.Key.Key_Right:
            # right arrow key
            self.next_roi()

    def closeEvent(self, event):
        retval = self.gui.exit_dialog()

        if retval == QMessageBox.StandardButton.Save:
            # Save before exit
            self.save_file()
            event.accept()
            self.data_handler.create_new_temp_hdf5_file()
        elif retval == QMessageBox.StandardButton.Discard:
            # Do not save before exit
            event.accept()
            self.data_handler.create_new_temp_hdf5_file()
        else:
            # Do not exit
            event.ignore()

    def mouse_moved(self, event):
        vb = self.gui.trace_plot_item.vb
        if self.gui.trace_plot_item.sceneBoundingRect().contains(event):
            mouse_point = vb.mapSceneToView(event)
            self.gui.mouse_label.setText(f"<p style='color:black'>X： {mouse_point.x():.4f} <br> Y: {mouse_point.y():.4f}</p>")



import json
import os
import math
import subprocess
import cv2
import configparser
from platformdirs import user_config_dir
from tifffile import imwrite
from PyQt6.QtCore import pyqtSignal, Qt
from PyQt6.QtWidgets import (
    QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QFileDialog,
    QLabel, QCheckBox, QComboBox, QApplication, QDoubleSpinBox, QLineEdit,
    QGroupBox, QFormLayout, QTextEdit
)
import ffmpy
from roibaview.plugins.base import BasePlugin


class VideoConverterPlugin(BasePlugin):
    name = "Video Converter"
    category = "utils"

    def __init__(self, config=None, parent=None):
        self.config = config
        self.parent = parent

    def apply(self, *_):
        self.window = VideoConverter()
        self.window.show()


class VideoConverter(QMainWindow):

    ffmpeg_dir_set = pyqtSignal()

    def __init__(self):
        super().__init__()
        CONFIG_FILENAME = "video_converter_settings.ini"
        config_dir = user_config_dir('roibaview')
        os.makedirs(config_dir, exist_ok=True)
        self.config_path = os.path.join(config_dir, CONFIG_FILENAME)
        # Create or read config
        self.config = configparser.ConfigParser()
        # If config file exists, read it
        if os.path.exists(self.config_path):
            self.config.read(self.config_path)
        else:
            # Create default config
            self.config['ffmpeg'] = {
                'dir': '',
            }
            with open(self.config_path, 'w') as configfile:
                self.config.write(configfile)

        self.input_file = None
        self.output_file = None

        self.crf_value = 17
        self.preset = 'superfast'
        self.output_frame_rate = 0

        self.use_gpu = False
        self.supress_terminal_output = True
        self.rescale_video_state = False

        central_widget = QWidget()
        layout = QVBoxLayout(central_widget)

        # Widgets
        self.browse_button = QPushButton("Browse Files")
        self.browse_button.clicked.connect(self.browse_files)

        self.input_file_label = QLabel('Input File not selected')
        self.output_file_label = QLabel('Output File not selected')

        self.start_button = QPushButton("Convert Video")
        self.start_button.clicked.connect(self.start_converting)

        self.extract_frames_button = QPushButton("Extract Frames")
        self.extract_frames_button.clicked.connect(self.start_extracting_frames)

        self.batch_button = QPushButton("Batch Processing ...")
        self.batch_button.clicked.connect(self.batch_processing)

        self.change_ffmpeg_dir_button = QPushButton("Set ffmpeg directory")
        self.change_ffmpeg_dir_button.clicked.connect(self.browse_file_ffmpeg)

        self.ffmpeg_dir_label = QLabel(f'ffmpeg at: {self.config["ffmpeg"]["dir"]}')

        self.gpu_check_box = QCheckBox()
        self.gpu_check_box.setCheckState(Qt.CheckState.Unchecked)
        self.gpu_check_box.stateChanged.connect(self.get_gpu_state)
        self.gpu_check_box_label = QLabel('Use GPU')

        self.supress_terminal_output_check_box = QCheckBox()
        self.supress_terminal_output_check_box.setCheckState(Qt.CheckState.Checked)
        self.supress_terminal_output_check_box.stateChanged.connect(self.get_supress_state)
        self.supress_terminal_output_label = QLabel('Suppress Terminal Output')

        self.quality_combo_box = QComboBox()
        self.quality_combo_box.addItems(['veryslow', 'slower', 'slow', 'medium', 'fast', 'faster', 'veryfast'])
        self.quality_combo_box_label = QLabel('Compression Preset')

        self.constant_rate_factor = QDoubleSpinBox()
        self.constant_rate_factor.setValue(self.crf_value)
        self.constant_rate_factor_label = QLabel('Video Quality: CRF (0-51, visually lossless=17)')

        self.change_frame_rate = QDoubleSpinBox()
        self.change_frame_rate.setValue(0)
        self.change_frame_rate_label = QLabel('Output Frame Rate (Hz, 0 to ignore)')

        self.rescale_check_box = QCheckBox()
        self.rescale_check_box.setCheckState(Qt.CheckState.Unchecked)
        self.rescale_check_box.stateChanged.connect(self.get_rescale_state)
        self.rescale_check_box_label = QLabel('Rescale Video')
        self.rescale_user_input = QLineEdit('WidthxHeight')

        self.status_label = QLabel('Ready')

        # --- Group Layouts ---
        io_group = QGroupBox("File Selection")
        io_layout = QVBoxLayout()
        io_layout.addWidget(self.browse_button)
        io_layout.addWidget(self.input_file_label)
        io_layout.addWidget(self.output_file_label)
        io_layout.addWidget(self.change_ffmpeg_dir_button)
        io_layout.addWidget(self.ffmpeg_dir_label)
        io_group.setLayout(io_layout)

        options_group = QGroupBox("Conversion Options")
        options_layout = QFormLayout()
        options_layout.addRow(self.gpu_check_box_label, self.gpu_check_box)
        options_layout.addRow(self.supress_terminal_output_label, self.supress_terminal_output_check_box)
        options_layout.addRow(self.quality_combo_box_label, self.quality_combo_box)
        options_layout.addRow(self.constant_rate_factor_label, self.constant_rate_factor)
        options_layout.addRow(self.change_frame_rate_label, self.change_frame_rate)
        rescale_layout = QHBoxLayout()
        rescale_layout.addWidget(self.rescale_check_box)
        rescale_layout.addWidget(self.rescale_user_input)
        options_layout.addRow(self.rescale_check_box_label, rescale_layout)
        options_group.setLayout(options_layout)

        actions_group = QGroupBox("Actions")
        actions_layout = QVBoxLayout()
        self.load_config_button = QPushButton("Load Config")
        self.save_config_button = QPushButton("Save Config")
        self.load_config_button.clicked.connect(self.load_config)
        self.save_config_button.clicked.connect(self.save_config)
        actions_layout.addWidget(self.start_button)
        actions_layout.addWidget(self.extract_frames_button)
        actions_layout.addWidget(self.batch_button)
        actions_layout.addWidget(self.load_config_button)
        actions_layout.addWidget(self.save_config_button)
        actions_group.setLayout(actions_layout)

        layout.addWidget(io_group)
        layout.addWidget(options_group)
        layout.addWidget(actions_group)
        layout.addWidget(self.status_label)

        # Logging area
        self.log_box = QTextEdit()
        self.log_box.setReadOnly(True)
        layout.addWidget(self.log_box)

        self.clear_log_button = QPushButton("Clear Log")
        self.clear_log_button.clicked.connect(self.log_box.clear)
        layout.addWidget(self.clear_log_button)

        self.setCentralWidget(central_widget)
        self.setWindowTitle("Video Converter")

        if self.config['ffmpeg']['dir'] == '':
            self.browse_file_ffmpeg()

        self.ffmpeg_dir = self.config['ffmpeg']['dir']
        self.ffmpeg_probe = f'{os.path.split(self.ffmpeg_dir)[0]}/ffprobe.exe'
        self._define_ffmpeg_settings()

    def log(self, message):
        self.log_box.append(message)

    def save_config(self):
        config_data = {
            "use_gpu": self.use_gpu,
            "suppress_output": self.supress_terminal_output,
            "preset": self.quality_combo_box.currentText(),
            "crf": self.constant_rate_factor.value(),
            "frame_rate": self.change_frame_rate.value(),
            "rescale": self.rescale_video_state,
            "rescale_size": self.rescale_user_input.text(),
        }
        file_name, _ = QFileDialog.getSaveFileName(self, "Save Config", "", "JSON Files (*.json)")
        if file_name:
            with open(file_name, 'w') as f:
                json.dump(config_data, f, indent=4)
            self.status_label.setText('Configuration saved.')
            self.log('Configuration saved.')

    def load_config(self):
        file_name, _ = QFileDialog.getOpenFileName(self, "Open Config", "", "JSON Files (*.json)")
        if file_name:
            with open(file_name, 'r') as f:
                config_data = json.load(f)
            self.gpu_check_box.setChecked(config_data.get("use_gpu", False))
            self.supress_terminal_output_check_box.setChecked(config_data.get("suppress_output", True))
            self.constant_rate_factor.setValue(config_data.get("crf", 17))
            self.change_frame_rate.setValue(config_data.get("frame_rate", 0))
            self.rescale_check_box.setChecked(config_data.get("rescale", False))
            self.rescale_user_input.setText(config_data.get("rescale_size", 'WidthxHeight'))
            preset = config_data.get("preset", "medium")
            index = self.quality_combo_box.findText(preset)
            if index != -1:
                self.quality_combo_box.setCurrentIndex(index)
            self.status_label.setText('Configuration loaded.')
            self.log('Configuration loaded.')

    def batch_processing(self):
        input_dir = QFileDialog.getExistingDirectory(self)
        if input_dir:
            # Get Files
            self.log("++++ BATCH PROCESSING ++++")
            self.input_file_label.setText("Batch Processing ...")
            file_list = os.listdir(input_dir)
            for f_name in file_list:
                self.input_file = f'{input_dir}/{f_name}'
                self.output_file = f'{input_dir}/{f_name[:-4]}_batch.mp4'
                self.log(f"++++ STARTING: {self.input_file} ++++")
                self._define_ffmpeg_settings()
                self.please_wait_status()
                QApplication.processEvents()
                self.convert_video(self.input_file, self.output_file)
                self.finished_status()

    def convert_to_tiff_stack(self):
        input_video = self.input_file
        output_tiff = self.output_file
        chunk_size = 300

        cap = cv2.VideoCapture(input_video)
        total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        num_chunks = math.ceil(total_frames / chunk_size)

        for chunk_idx, start_frame in enumerate(range(0, total_frames, chunk_size)):
            frames = []
            cap_chunk = cv2.VideoCapture(input_video)
            cap_chunk.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
            for _ in range(chunk_size):
                ret, frame = cap_chunk.read()
                if ret:
                    frames.append(frame)
                else:
                    break

            imwrite(output_tiff, frames, append=True, bigtiff=True, compression='deflate')
            self.log(f"Processed chunk {chunk_idx + 1}/{num_chunks}")

            cap_chunk.release()  # Release the video capture object

        cap.release()  # Release the main video capture object

    def _define_ffmpeg_settings(self):
        self.ffmpeg_input_opt = {'gpu': ['-hwaccel', 'cuda', '-hwaccel_output_format', 'cuda'], 'cpu': None}

        # Get Values
        self.crf_value = int(self.constant_rate_factor.value())
        self.preset = self.quality_combo_box.currentText()
        self.output_frame_rate = int(self.change_frame_rate.value())

        if self.crf_value > 51:
            self.crf_value = 51

        # First define the basic settings
        self.ffmpeg_output_opt = {
            'gpu': ['-c:v', 'h264_nvenc', '-preset', self.preset, '-qp', str(self.crf_value)],
            'cpu': ['-c:v', 'libx264', '-preset', self.preset, '-crf', str(self.crf_value)],
            'avi': ['-an', '-vcodec', 'rawvideo', '-y']
        }

        # Now add more settings if desired
        # Change frame rate
        if self.output_frame_rate > 0:
            self.ffmpeg_output_opt['gpu'].extend(['-filter:v', f'fps={self.output_frame_rate}'])
            self.ffmpeg_output_opt['cpu'].extend(['-filter:v', f'fps={self.output_frame_rate}'])
            self.ffmpeg_output_opt['avi'].extend(['-filter:v', f'fps={self.output_frame_rate}'])

        # Change size
        if self.rescale_video_state:
            # Get Video Size in Pixel
            video_size = self.rescale_user_input.text()
            video_size = video_size.split('x')
            if len(video_size) == 2:
                self.log(f'New Video Size: {video_size[0]} x {video_size[1]} pixel')
            else:
                self.log('Input incorrect!')
                return None
            self.ffmpeg_output_opt['gpu'].extend(['-vf', f'scale_cuda={video_size[0]}:{video_size[1]}'])
            self.ffmpeg_output_opt['cpu'].extend(['-vf', f'scale={video_size[0]}:{video_size[1]}'])
            self.ffmpeg_output_opt['avi'].extend(['-vf', f'scale={video_size[0]}:{video_size[1]}'])

        self.ffmpeg_global_opt = {
            'supress': ['-y', '-loglevel', 'quiet'],
            'show': ['-y'],
        }

    def get_rescale_state(self):
        if self.rescale_check_box.checkState() == Qt.CheckState.Checked:
            self.rescale_video_state = True
        else:
            self.rescale_video_state = False

    def get_gpu_state(self):
        if self.gpu_check_box.checkState() == Qt.CheckState.Checked:
            self.use_gpu = True
            # change presets
            self.quality_combo_box.clear()
            self.quality_combo_box.addItem('slow')
            self.quality_combo_box.addItem('medium')
            self.quality_combo_box.addItem('fast')
        else:
            self.use_gpu = False
            # change presets
            self.quality_combo_box.clear()
            self.quality_combo_box.addItem('veryslow')
            self.quality_combo_box.addItem('slower')
            self.quality_combo_box.addItem('slow')
            self.quality_combo_box.addItem('medium')
            self.quality_combo_box.addItem('fast')
            self.quality_combo_box.addItem('faster')
            self.quality_combo_box.addItem('veryfast')

    def get_supress_state(self):
        if self.supress_terminal_output_check_box.checkState() == Qt.CheckState.Checked:
            self.supress_terminal_output = True
        else:
            self.supress_terminal_output = False

    def get_video_info(self, filename):
        result = subprocess.run([self.ffmpeg_probe, "-v", "error", "-select_streams", "v:0", "-show_entries",
                                 "stream=duration:stream=avg_frame_rate", "-of",
                                 "default=noprint_wrappers=1:nokey=1", filename],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        output = result.stdout.decode("utf-8").split('\r\n')
        frame_rate = float(output[0].split('/')[0])  # Extracting the first part before '/'
        duration = float(output[1])

        return frame_rate, duration

    def extract_frames(self, input_file, save_dir):
        if self.ffmpeg_dir is not None:
            # ffmpeg -i input.mp4 -vf fps=1 %04d.png
            video_frame_rate, video_duration = self.get_video_info(input_file)

            if self.output_frame_rate > 0:
                output_cmd = ['-vf', f'fps={self.output_frame_rate}']
                fr = self.output_frame_rate
            else:
                output_cmd = None
                fr = video_frame_rate

            number_of_frames = int(fr * video_duration)
            self.log(f'++++ Expecting to Store {number_of_frames} Frames to HDD+++')
            counter = len(str(number_of_frames)) + 1
            output_dir = f'{save_dir}/%0{counter}d.jpg'
            if self.supress_terminal_output:
                global_settings = self.ffmpeg_global_opt['supress']
            else:
                global_settings = self.ffmpeg_global_opt['show']

            ff = ffmpy.FFmpeg(
                executable=self.ffmpeg_dir,
                global_options=global_settings,
                inputs={input_file: None},
                outputs={output_dir: output_cmd}
            )
            ff.run()

    def start_extracting_frames(self):
        file_dir = str(QFileDialog.getExistingDirectory(self, "Select Directory"))
        if self.input_file is not None and file_dir is not None:
            self.log('START EXTRACTING')
            # self._define_ffmpeg_settings()
            self.output_frame_rate = int(self.change_frame_rate.value())
            self.please_wait_status()
            QApplication.processEvents()
            self.extract_frames(self.input_file, file_dir)
            self.finished_status()

    def convert_video(self, input_file, output_file):
        if self.ffmpeg_dir is not None:
            if output_file[-3:] == 'tif':
                self.log('CONVERT TO TIFF')
                self.convert_to_tiff_stack()
            else:
                # check settings
                if self.use_gpu:
                    hw = 'gpu'
                else:
                    hw = 'cpu'

                if output_file[-3:] == 'avi':
                    # Use no compression for avi file (otherwise you can not open it in imagej)
                    input_cmd = self.ffmpeg_input_opt['cpu']
                    output_cmd = self.ffmpeg_output_opt['avi']
                else:
                    input_cmd = self.ffmpeg_input_opt[hw]
                    output_cmd = self.ffmpeg_output_opt[hw]

                if self.supress_terminal_output:
                    global_settings = self.ffmpeg_global_opt['supress']
                else:
                    global_settings = self.ffmpeg_global_opt['show']
                    # Print ffmpeg settings to terminal
                    self.log("FFMPEG COMMAND:")
                    self.log(f'INPUT: {input_cmd}')
                    self.log(f'OUTPUT: {output_cmd}')

                ff = ffmpy.FFmpeg(
                    executable=self.ffmpeg_dir,
                    global_options=global_settings,
                    inputs={input_file: input_cmd},
                    outputs={output_file: output_cmd}
                )
                if self.supress_terminal_output:
                    out, err = ff.run(stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    self.log(out.decode('utf-8'))
                    self.log(err.decode('utf-8'))
                else:
                    ff.run()

    def browse_file_ffmpeg(self):
        self.ffmpeg_dir, _ = QFileDialog.getOpenFileName(self, "Select FFMPEG .exe", "", "ffmpeg (*.exe)")
        # Modify and save
        self.config['ffmpeg']['dir'] = self.ffmpeg_dir
        with open(self.config_path, 'w') as configfile:
            self.config.write(configfile)

        self.ffmpeg_dir_label.setText(f'ffmpeg at: {self.ffmpeg_dir}')

    def please_wait_status(self):
        self.status_label.setText('Please wait ... ')
        self.log('Please wait ...')
        self.browse_button.setDisabled(True)
        self.change_ffmpeg_dir_button.setDisabled(True)

    def finished_status(self):
        self.status_label.setText('Converting finished!')
        self.browse_button.setDisabled(False)
        self.change_ffmpeg_dir_button.setDisabled(False)
        if not self.supress_terminal_output:
            self.log(f"++++ FINISHED: {os.path.split(self.input_file)[1]} --> {os.path.split(self.output_file)[1]} ++++")

    def browse_files(self):
        input_file, _ = QFileDialog.getOpenFileName(
            self, "Select Input File", "", "Video Files (*.mp4; *.avi; *.mkv; *.mpeg; *.mpg)")
        if input_file:
            self.input_file = input_file
            self.input_file_label.setText(input_file)

    def start_converting(self):
        output_file, _ = QFileDialog.getSaveFileName(
            self, "Select Output File", "", "MP4, (*.mp4);; AVI, (*.avi);; MKV, (*.mkv);; TIF, (*.tif)")
        if output_file:
            self.output_file = output_file
            self.output_file_label.setText(output_file)

        if self.input_file is not None and self.output_file is not None:
            self._define_ffmpeg_settings()
            self.log('==== START CONVERTING VIDEO ====')
            self.please_wait_status()
            QApplication.processEvents()
            self.convert_video(self.input_file, self.output_file)
            self.finished_status()



from setuptools import setup, find_packages

setup(
    name="roibaview",
    version="0.1.8",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "PyQt6",
        "pyqtgraph",
        "numpy",
        "pandas",
        "scipy",
        "h5py",
        "tifftools",
        "tifffile",
        "opencv-python",
        "ffmpy",
        "joblib",
        "platformdirs",
        "roifile"
    ],
    entry_points={
        "gui_scripts": [
            "roibaview = roibaview.main:main",
        ],
    },
    python_requires=">=3.7",
    author="Nils",
    author_email="your.email@example.com",
    description="A PyQt6-based application for ROI visualization.",
    url="https://github.com/UniNilsBrehm/roibaview",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)


import pyqtgraph as pg
import os
import numpy as np
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import QVBoxLayout, QLabel, QSpacerItem, QSizePolicy, QMessageBox, QDialog, \
    QPushButton, QLineEdit
from roibaview.gui_utils import BrowseFileDialog
import pandas as pd
import scipy.signal as sig
from roibaview.plugins.base import BasePlugin


class VentralRootDetectionPlugin(BasePlugin):
    name = "Ventral Root Detection"
    category = "tool"

    def __init__(self, config=None, parent=None):
        self.config = config
        self.parent = parent
        self.roi_signal = getattr(parent, "signal_roi_idx_changed", None)
        self.dialog = None

    def apply(self, *_):
        try:
            controller = getattr(self.parent, "controller", None)
            if not controller or not controller.selected_data_sets:
                raise RuntimeError("No controller or dataset selected.")

            name = controller.selected_data_sets[0]
            dtype = controller.selected_data_sets_type[0]
            roi = controller.current_roi_idx
            data = controller.data_handler.get_data_set(dtype, name)
            meta = controller.data_handler.get_data_set_meta_data(dtype, name)
            fr = meta["sampling_rate"]
            master_plot = controller.data_plotter.master_plot

            self.dialog = VentralRootDetection(
                data=data + meta["y_offset"],
                fr=fr,
                master_plot=master_plot,
                roi=roi,
                parent=self.parent
            )

            if self.roi_signal:
                self.roi_signal.connect(self.dialog.roi_changed)

            self.dialog.finished.connect(self._on_dialog_closed)
            self.dialog.show()

        except Exception as e:
            QMessageBox.critical(self.parent, "Ventral Root Detection Error", str(e))

    def _on_dialog_closed(self):
        if self.roi_signal and hasattr(self, 'dialog'):
            try:
                self.roi_signal.disconnect(self.dialog.roi_changed)
            except Exception:
                pass
        if hasattr(self, 'dialog') and self.dialog:
            self.dialog.deleteLater()
            self.dialog = None


class VentralRootDetection(QDialog):
    signal_roi_changed = pyqtSignal(int)
    main_window_closing = pyqtSignal()
    def __init__(self, data, fr, master_plot, roi, parent=None, **kwargs):
        super().__init__(parent)
        self.roi_idx = roi
        self.data = data  # this is the data set
        self.data_trace = self.data[:, self.roi_idx]  # this is the roi trace
        self.fr = fr
        self.time_axis = self.compute_time_axis(self.data_trace.shape[0], self.fr)
        self.master_plot = master_plot
        self.parameters = dict()
        self.parameters_range = dict()
        self.vr_events = dict()
        self.env_trace = None
        self.set_parameters()
        self.min_range = 0
        self.max_range = 1000
        # self.set_parameters_range()
        self._init_ui()
        self.main_window_running = True
        self.signal_roi_changed.connect(self.roi_changed)
        self.main_window_closing.connect(self.main_window_is_closing)

    def export_peaks(self):
        file_browser = BrowseFileDialog(self)
        file_dir = file_browser.save_file_name('csv file, (*.csv)')
        if file_dir:
            result = pd.DataFrame()
            result['onset_time'] = self.vr_events['onset_times']
            result['offset_time'] = self.vr_events['offset_times']
            result['onset_idx'] = self.vr_events['onset_idx']
            result['offset_idx'] = self.vr_events['offset_idx']
            result.to_csv(file_dir, index=False)
            env = pd.DataFrame(self.env_trace)
            env_dir = os.path.split(file_dir)[0] + '/envelope_trace.csv'
            env.to_csv(env_dir, index=False, header=False)

    @staticmethod
    def compute_time_axis(data_size, fr):
        max_time = data_size / fr
        return np.linspace(0, max_time, data_size)

    def main_window_is_closing(self):
        self.main_window_running = False
        self.close()

    def roi_changed(self, new_roi_idx):
        print('ROI CHANGED')
        self.roi_idx = new_roi_idx
        self.data_trace = self.data[:, self.roi_idx]

        # Update find vr events
        self.find_vr_events(self.parameters)
        self.update_plot()

    def _init_ui(self):
        layout = QVBoxLayout()
        self.parameters_labels = dict()
        for param_name in self.parameters.keys():
            label = QLabel(param_name)
            # Add text field
            field = QLineEdit()

            self.parameters_labels[param_name] = QLabel('0')

            # Connect slider signal to update function
            # slider.valueChanged.connect(lambda value, param=param_name: self.update_parameter(param, value))
            field.textChanged.connect(lambda value, param=param_name: self.update_parameter_only(param, value))
            layout.addWidget(label)
            # layout.addWidget(slider)
            layout.addWidget(field)
            layout.addWidget(self.parameters_labels[param_name])
            # Add spacer (width, height)
            spacer = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
            layout.addItem(spacer)

        self.update_button = QPushButton('UPDATE')
        self.update_button.clicked.connect(self.update_detection)
        layout.addWidget(self.update_button)

        self.export_button = QPushButton('Export...')
        self.export_button.clicked.connect(self.export_peaks)
        layout.addWidget(self.export_button)

        self.setLayout(layout)
        self.setWindowTitle("Peak Detection Parameters")
        # self.show()

    def update_detection(self):
        # Update Value Label
        for param_name in self.parameters.keys():
            self.parameters_labels[param_name].setText(f'{self.parameters[param_name]:.2f}')

        # Update find vr events
        self.find_vr_events(self.parameters)

        # Update Plot
        self.update_plot()

    def update_parameter_only(self, param_name, value):
        try:
            val = float(value)
        except ValueError:
            val = 1

        # Update parameter value
        self.parameters[param_name] = val

    def update_parameter(self, param_name, value):
        # Update parameter value
        if value == 0:
            self.parameters[param_name] = None
        else:
            self.parameters[param_name] = value

        parameters = self.parameters
        if value is not None:
            parameters[param_name] = value
        else:
            parameters[param_name] = None

        # Update Value Label
        self.parameters_labels[param_name].setText(f'{parameters[param_name]:.2f}')

        # Update find vr events
        self.find_vr_events(parameters)

        # Update Plot
        self.update_plot()

    def clear_plot(self):
        # check if there is already roi data plotted and remove it
        item_list = self.master_plot.items.copy()
        for item in item_list:
            if item.name() is not None:
                if item.name().startswith('event'):
                    self.master_plot.removeItem(item)

        for item in item_list:
            if item.name() is not None:
                if item.name().startswith('env'):
                    self.master_plot.removeItem(item)

    def update_plot(self):
        # Check if there is already roi data plotted and remove it
        self.clear_plot()

        if self.env_trace is not None:
            # Plot the Envelope
            plot_data_item_env = pg.PlotDataItem(
                self.time_axis, self.env_trace,
                pen=pg.mkPen(color=(0, 255, 0)),
                name='env',
                skipFiniteCheck=True,
                tip=None,
            )
            self.master_plot.addItem(plot_data_item_env)

            # Plot event onsets/offsets only if present
            if 'onset_idx' in self.vr_events and 'offset_idx' in self.vr_events:
                if len(self.vr_events['onset_idx']) > 0:
                    onset_plot = pg.ScatterPlotItem(
                        self.time_axis[self.vr_events['onset_idx']], self.env_trace[self.vr_events['onset_idx']],
                        pen=pg.mkPen(color=(255, 0, 0)),
                        brush=pg.mkBrush(color=(255, 0, 0)),
                        size=50,
                        symbol='arrow_down',
                        name='event_onset',
                        skipFiniteCheck=True,
                        tip=None,
                        hoverable=True,
                        hoverSize=100
                    )
                    self.master_plot.addItem(onset_plot)

                if len(self.vr_events['offset_idx']) > 0:
                    offset_plot = pg.ScatterPlotItem(
                        self.time_axis[self.vr_events['offset_idx']], self.env_trace[self.vr_events['offset_idx']],
                        pen=pg.mkPen(color=(0, 0, 255)),
                        brush=pg.mkBrush(color=(0, 0, 255)),
                        size=50,
                        symbol='arrow_down',
                        name='event_offset',
                        skipFiniteCheck=True,
                        tip=None,
                        hoverable=True,
                        hoverSize=100
                    )
                    self.master_plot.addItem(offset_plot)

    def set_parameters(self):
        self.parameters['threshold'] = 5  # times std
        self.parameters['large_th_factor'] = 20
        # self.parameters['vr_fr'] = self.fr  # in Hz
        self.parameters['vr_cutoff'] = 5  # in Hz
        self.parameters['movingaverage_window'] = 2  # in samples
        self.parameters['duration_th_secs'] = 5  # in secs. Events longer than that will be ignored
        self.parameters['minimal_event_distance'] = 4  # in secs. Events with a distance of less that will be merged

    @staticmethod
    def map_range(value, from_min, from_max, to_min, to_max):
        # First, scale the value from the input range to a 0-1 range
        scaled_value = (value - from_min) / (from_max - from_min)

        # Then, scale the value to fit within the output range
        mapped_value = to_min + (to_max - to_min) * scaled_value

        return mapped_value

    @staticmethod
    def envelope(data, rate, freq):
        nyquist = rate / 2
        if freq >= nyquist:
            freq = nyquist * 0.99  # Clamp just below Nyquist
        if freq <= 0:
            freq = nyquist * 0.1  # Ensure it's > 0

        sos = sig.butter(2, freq, 'lowpass', fs=rate, output='sos')
        env = (np.sqrt(2) * sig.sosfiltfilt(sos, np.abs(data))) ** 2
        return env

    @staticmethod
    def moving_average_filter(data, window):
        return np.convolve(data, np.ones(int(window)) / int(window), mode='same')

    @staticmethod
    def z_transform(data):
        result = (data - np.mean(data, axis=0)) / np.std(data, axis=0)
        return result

    def find_vr_events(self, parameters):
        data = self.data_trace.copy()

        if parameters['threshold'] is None:
            # If there is no Detection Threshold set one like this
            # Standard Deviation of the trace times 5
            th = np.std(data) * 5
        else:
            th = np.std(data) * parameters['threshold']

        # Compute envelope
        env = self.envelope(data=data, rate=self.fr, freq=parameters['vr_cutoff'])

        # Low pass filter ventral root envelope with a moving average
        env_fil = self.moving_average_filter(env, window=parameters['movingaverage_window'])
        self.env_trace = env_fil.copy()
        env_vr_z = self.z_transform(env_fil)

        # Create Binary
        binary = np.zeros_like(env_vr_z)
        binary[env_vr_z > th] = 1

        # Find onsets and offsets of ventral root activity
        onsets_offsets = np.diff(binary, append=0)
        time_axis = self.time_axis
        onset_idx = np.where(onsets_offsets > 0)[0]
        offset_idx = np.where(onsets_offsets < 0)[0]
        onset_times = time_axis[onset_idx]
        offset_times = time_axis[offset_idx]

        # check for motor activity that is too long (artifacts due to concatenating recordings) and remove it
        if len(offset_times) == len(onset_times):
            event_duration = offset_times - onset_times
            idx_remove = event_duration > parameters['duration_th_secs']
            onset_times = onset_times[np.invert(idx_remove)]
            offset_times = offset_times[np.invert(idx_remove)]
        else:
            print('WARNING: Number of onset times and offset times do not match!')

        # Merge Events that are to close to each other
        # Combine event information into a single list of tuples
        events = list(zip(onset_times, offset_times, onset_idx, offset_idx))
        # Sort events by their onset time
        events = sorted(events, key=lambda x: x[0])

        # Initialize lists to store merged events and their indices
        merged_onset_indices = []
        merged_offset_indices = []
        merged_onset_times = []
        merged_offset_times = []

        if not events:
            # No events detected, clear result and exit early
            self.vr_events.clear()
            return
        # Initialize the first event as the current event to merge
        current_onset_time, current_offset_time, current_onset_index, current_offset_index = events[0]

        for onset_time, offset_time, onset_index, offset_index in events[1:]:
            # Check if the current event is close to the previous one
            if onset_time - current_offset_time <= parameters['minimal_event_distance']:
                # Merge the events by extending the offset time and updating indices
                current_offset_time = max(current_offset_time, offset_time)
                current_offset_index = offset_index
            else:
                # If not close, save the current event and start a new one
                merged_onset_times.append(current_onset_time)
                merged_offset_times.append(current_offset_time)
                merged_onset_indices.append(current_onset_index)
                merged_offset_indices.append(current_offset_index)
                current_onset_time, current_offset_time = onset_time, offset_time
                current_onset_index, current_offset_index = onset_index, offset_index

        # Append the last merged event
        merged_onset_times.append(current_onset_time)
        merged_offset_times.append(current_offset_time)
        merged_onset_indices.append(current_onset_index)
        merged_offset_indices.append(current_offset_index)

        # Convert lists to numpy arrays
        merged_onset_times = np.array(merged_onset_times)
        merged_offset_times = np.array(merged_offset_times)
        merged_onset_indices = np.array(merged_onset_indices)
        merged_offset_indices = np.array(merged_offset_indices)

        # Collect results
        self.vr_events['onset_idx'] = merged_onset_indices
        self.vr_events['offset_idx'] = merged_offset_indices
        self.vr_events['onset_times'] = merged_onset_times
        self.vr_events['offset_times'] = merged_offset_times

    def closeEvent(self, event):
        self._cleanup()
        self.done(QDialog.DialogCode.Accepted)
        event.accept()
        # import gc
        # gc.collect()
        # print("Remaining VentralRootDetection instances:",
        #       [o for o in gc.get_objects() if isinstance(o, VentralRootDetection)])

    def _cleanup(self):
        try:
            if self.parent() and hasattr(self.parent(), "signal_roi_idx_changed"):
                self.parent().signal_roi_idx_changed.disconnect(self.roi_changed)
        except Exception:
            pass
        self.clear_plot()

    # def __del__(self):
    #     print("VentralRootDetection dialog deleted.")



import h5py
import os
import shutil
import numpy as np
import pandas as pd
import tempfile
from PyQt6.QtCore import pyqtSignal, QObject
from roibaview.gui_utils import MessageBox
"""
Notes:
    - Try to use PyTables instead of h5py
    - Try using NixPy

Data Structure:
.
├── data_sets
│   ├── set_01
│   ├── set_02
:   :
│   └── set_n
│
├── global_data_sets
│   ├── global_set_01
:   :
│   └── global_set_n
│
└── some_other_stuff
    ├── stuff_01
    └── info
"""


class DataHandler(QObject):
    signal_roi_id_changed = pyqtSignal()

    def __init__(self):
        QObject.__init__(self)
        self._set_csv_import_settings()
        temp_dir = tempfile.gettempdir()
        self.temp_file_name = os.path.join(temp_dir, "temp_data.hdf5")
        self.create_new_temp_hdf5_file()
        self.roi_count = 0

    def _set_csv_import_settings(self):
        # Settings for importing a csv file using pandas
        # The decimal symbol (english: ".", german: ",")
        self.csv_decimal = '.'

        # The separation/delimiter symbol: "," or "\t" or ";" or etc.
        self.csv_sep = ','

    def create_new_temp_hdf5_file(self):
        # Will create an empty hdf5 file into the temp directory with one group called "data_sets"
        with h5py.File(self.temp_file_name, 'w') as f:
            f.create_group('data_sets')
            f.create_group('global_data_sets')

    def import_csv(self, file_dir, data_name, sampling_rate, data_set_type, column):
        # The .csv file: Each Column is the data of one ROI so the shape is (Samples, ROIs)
        # First check if there are headers (ROI Names)
        data_check = pd.read_csv(file_dir, decimal=self.csv_decimal, sep=self.csv_sep, index_col=None, nrows=1, header=None)
        if data_check.iloc[0, :].dtype == 'O':
            # There are headers
            data_file = pd.read_csv(file_dir, decimal=self.csv_decimal, sep=self.csv_sep, index_col=None)

            # Remove Index Column if there is one
            if 'Unnamed: 0' in data_file.keys():
                data_file.drop(columns='Unnamed: 0', inplace=True)
            headers = np.array(data_file.keys())
        else:
            data_file = pd.read_csv(file_dir, decimal=self.csv_decimal, sep=self.csv_sep, index_col=None, header=None)

            # Remove Index Column if there is one
            if 'Unnamed: 0' in data_file.keys():
                data_file.drop(columns='Unnamed: 0', inplace=True)

            headers = np.arange(0, data_file.shape[1], 1)

        # This needs to be converted to a numpy matrix
        data = data_file.to_numpy()

        # Check if user wants only a specific column
        try:
            col_id = int(column)
        except ValueError:
            col_id = None

        if col_id is not None and data_set_type == 'global_data_sets':
            try:
                data = data[:, col_id].reshape(-1, 1)
            except IndexError:
                MessageBox(title='ERROR', text='Selected column exceeds number of columns in data!')
                return False

        # Check dimensions (must match hdf5 style)
        if data.ndim == 1:
            data = np.atleast_2d(data)

        # Open the temp hdf5 file and store data set there
        check = self.add_new_data_set(
            data_set_type, data_name, data,
            sampling_rate=sampling_rate,
            time_offset=0,
            y_offset=0,
            header=headers)
        return True

    def get_info(self):
        with h5py.File(self.temp_file_name, 'r') as f:
            groups = list(f.keys())
            results = dict()
            for gr in groups:
                results[gr] = list(f[gr].keys())
        return results

    def check_if_exists(self, data_set_type, data_set_name):
        with h5py.File(self.temp_file_name, 'r') as f:
            if data_set_name in f[data_set_type]:
                print('ERROR: Data set with this name already exists!')
                return True
            else:
                return False

    def delete_column(self, data_set_type, data_set_name, col_nr):
        with h5py.File(self.temp_file_name, 'r+') as f:
            if data_set_name in f[data_set_type]:
                # Convert to a NumPy array
                dset = f[data_set_type][data_set_name]
                data = dset[:]

                # Remove the column
                try:
                    modified_data = np.delete(data, col_nr, axis=1)

                    # Resize the dataset to match new shape
                    dset.resize(modified_data.shape)

                    # Overwrite dataset with new data
                    dset[...] = modified_data  # Overwrite without deleting the dataset
                    self.roi_count = modified_data.shape[1]

                except IndexError:
                    return None

    def delete_data_set(self, data_set_type, data_set_name):
        with h5py.File(self.temp_file_name, 'r+') as f:
            if data_set_name in f[data_set_type]:
                del f[data_set_type][data_set_name]

    def rename_data_set(self, data_set_type, data_set_name, new_name):
        with h5py.File(self.temp_file_name, 'r+') as f:
            if data_set_name in f[data_set_type]:
                f[data_set_type][new_name] = f[data_set_type][data_set_name]
                del f[data_set_type][data_set_name]

    def add_new_data_set(self, data_set_type, data_set_name, data, sampling_rate, time_offset, y_offset, header):
        # Open the temp hdf5 file and store data set there
        already_exists = False
        with h5py.File(self.temp_file_name, 'r+') as f:
            # Check if data set is available
            if data_set_name in f[data_set_type]:
                MessageBox(title='ERROR', text='Data set with this name already exists!')
                data_set_name = data_set_name + '_new'
                already_exists = True

            # CREATE NEW DATASET
            new_entry = f[data_set_type].create_dataset(data_set_name, data=data, chunks=True)
            if data_set_type == 'global_data_sets':
                header_name = 'header_names'
            else:
                header_name = 'roi_names'
                self.roi_count = data.shape[1]
            new_entry.attrs[header_name] = header
            new_entry.attrs['sampling_rate'] = float(sampling_rate)
            new_entry.attrs['time_offset'] = time_offset
            new_entry.attrs['y_offset'] = y_offset
            new_entry.attrs['color'] = '#000000'  # black
            new_entry.attrs['lw'] = 1
            new_entry.attrs['name'] = data_set_name
            new_entry.attrs['data_type'] = data_set_type

            return already_exists

    def add_meta_data(self, data_set_type, data_set_name, metadata_dict):
        with h5py.File(self.temp_file_name, 'r+') as f:
            # Check if data set is available
            if data_set_name in f[data_set_type]:
                for k in metadata_dict:
                    data_set = f[data_set_type][data_set_name]
                    data_set.attrs[k] = metadata_dict[k]
            else:
                print('ERROR: Data set not found!')

    def get_data_set_meta_data(self, data_set_type, data_set_name):
        with h5py.File(self.temp_file_name, 'r') as f:
            # Check if data set is available
            if data_set_name in f[data_set_type]:
                data_set = f[data_set_type][data_set_name]
                meta_data = dict(data_set.attrs)
                return meta_data
            else:
                print('ERROR: Data set not found!')
                return None

    def get_data_set(self, data_set_type, data_set_name):
        # Get a specific data set
        # If it is available store it into a numpy array (RAM) and return it
        with h5py.File(self.temp_file_name, 'r') as f:
            # Check if data set is available
            if data_set_name in f[data_set_type]:
                data_set = f[data_set_type][data_set_name][:]
                return data_set
            else:
                print('ERROR: Data set not found!')
                return None

    def get_roi_data(self, data_set_name, roi_idx):
        # Get the data for a specific ROI in a specific data set
        # If it is available store it into a numpy array (RAM) and return it
        with h5py.File(self.temp_file_name, 'r') as f:
            # Check if data set is available
            if data_set_name in f['data_sets']:
                data_set = f['data_sets'][data_set_name]
            else:
                print('ERROR: Data set not found!')
                return None
            # Check if roi idx is in data set
            if data_set.shape[1] > roi_idx:
                roi_data = data_set[:, roi_idx]
                return roi_data
            else:
                print('ERROR: ROI Index is outside of data set range!')
                return None

    def save_file(self, file_dir):
        shutil.copyfile(self.temp_file_name, file_dir)

    def open_file(self, file_dir):
        shutil.copyfile(file_dir, self.temp_file_name)

    def new_file(self):
        self.create_new_temp_hdf5_file()

    def get_roi_count(self, data_set_name):
        with h5py.File(self.temp_file_name, 'r') as f:
            # Check if data set is available
            if data_set_name in f['data_sets']:
                # Cols = Rois
                roi_count = f['data_sets'][data_set_name].shape[1]
                return roi_count
            else:
                print('ERROR: Data set not found!')
                return None

    @staticmethod
    def compute_time_axis(data_size, fr):
        max_time = data_size / fr
        return np.linspace(0, max_time, data_size)
